import socket
import types
from unittest import mock
from unittest.mock import patch

import pytest
import valkey
from valkey import ConnectionPool, Valkey
from valkey._parsers import _LibvalkeyParser, _RESP2Parser, _RESP3Parser
from valkey.backoff import NoBackoff
from valkey.connection import (
    Connection,
    SSLConnection,
    UnixDomainSocketConnection,
    parse_url,
)
from valkey.exceptions import ConnectionError, InvalidResponse, TimeoutError
from valkey.retry import Retry
from valkey.utils import LIBVALKEY_AVAILABLE

from .conftest import skip_if_server_version_lt
from .mocks import MockSocket


@pytest.mark.skipif(LIBVALKEY_AVAILABLE, reason="PythonParser only")
@pytest.mark.onlynoncluster
def test_invalid_response(r):
    raw = b"x"
    parser = r.connection._parser
    with mock.patch.object(parser._buffer, "readline", return_value=raw):
        with pytest.raises(InvalidResponse) as cm:
            parser.read_response()
    assert str(cm.value) == f"Protocol Error: {raw!r}"


@skip_if_server_version_lt("4.0.0")
@pytest.mark.valkeymod
def test_loading_external_modules(r):
    def inner():
        pass

    r.load_external_module("myfuncname", inner)
    assert getattr(r, "myfuncname") == inner
    assert isinstance(getattr(r, "myfuncname"), types.FunctionType)

    # and call it
    from valkey.commands import ValkeyModuleCommands

    j = ValkeyModuleCommands.json
    r.load_external_module("sometestfuncname", j)

    # d = {'hello': 'world!'}
    # mod = j(r)
    # mod.set("fookey", ".", d)
    # assert mod.get('fookey') == d


class TestConnection:
    def test_disconnect(self):
        conn = Connection()
        mock_sock = mock.Mock()
        conn._sock = mock_sock
        conn.disconnect()
        mock_sock.shutdown.assert_called_once()
        mock_sock.close.assert_called_once()
        assert conn._sock is None

    def test_disconnect__shutdown_OSError(self):
        """An OSError on socket shutdown will still close the socket."""
        conn = Connection()
        mock_sock = mock.Mock()
        conn._sock = mock_sock
        conn._sock.shutdown.side_effect = OSError
        conn.disconnect()
        mock_sock.shutdown.assert_called_once()
        mock_sock.close.assert_called_once()
        assert conn._sock is None

    def test_disconnect__close_OSError(self):
        """An OSError on socket close will still clear out the socket."""
        conn = Connection()
        mock_sock = mock.Mock()
        conn._sock = mock_sock
        conn._sock.close.side_effect = OSError
        conn.disconnect()
        mock_sock.shutdown.assert_called_once()
        mock_sock.close.assert_called_once()
        assert conn._sock is None

    def clear(self, conn):
        conn.retry_on_error.clear()

    def test_retry_connect_on_timeout_error(self):
        """Test that the _connect function is retried in case of a timeout"""
        conn = Connection(retry_on_timeout=True, retry=Retry(NoBackoff(), 3))
        origin_connect = conn._connect
        conn._connect = mock.Mock()

        def mock_connect():
            # connect only on the last retry
            if conn._connect.call_count <= 2:
                raise socket.timeout
            else:
                return origin_connect()

        conn._connect.side_effect = mock_connect
        conn.connect()
        assert conn._connect.call_count == 3
        self.clear(conn)

    def test_connect_without_retry_on_os_error(self):
        """Test that the _connect function is not being retried in case of a OSError"""
        with patch.object(Connection, "_connect") as _connect:
            _connect.side_effect = OSError("")
            conn = Connection(retry_on_timeout=True, retry=Retry(NoBackoff(), 2))
            with pytest.raises(ConnectionError):
                conn.connect()
            assert _connect.call_count == 1
            self.clear(conn)

    def test_connect_timeout_error_without_retry(self):
        """Test that the _connect function is not being retried if retry_on_timeout is
        set to False"""
        conn = Connection(retry_on_timeout=False)
        conn._connect = mock.Mock()
        conn._connect.side_effect = socket.timeout

        with pytest.raises(TimeoutError) as e:
            conn.connect()
        assert conn._connect.call_count == 1
        assert str(e.value) == "Timeout connecting to server"
        self.clear(conn)


@pytest.mark.onlynoncluster
@pytest.mark.parametrize(
    "parser_class",
    [_RESP2Parser, _RESP3Parser, _LibvalkeyParser],
    ids=["RESP2Parser", "RESP3Parser", "LibvalkeyParser"],
)
def test_connection_parse_response_resume(r: valkey.Valkey, parser_class):
    """
    This test verifies that the Connection parser,
    be that PythonParser or LibvalkeyParser,
    can be interrupted at IO time and then resume parsing.
    """
    if parser_class is _LibvalkeyParser and not LIBVALKEY_AVAILABLE:
        pytest.skip("Libvalkey not available)")
    args = dict(r.connection_pool.connection_kwargs)
    args["parser_class"] = parser_class
    conn = Connection(**args)
    conn.connect()
    message = (
        b"*3\r\n$7\r\nmessage\r\n$8\r\nchannel1\r\n"
        b"$25\r\nhi\r\nthere\r\n+how\r\nare\r\nyou\r\n"
    )
    mock_socket = MockSocket(message, interrupt_every=2)

    if isinstance(conn._parser, _RESP2Parser) or isinstance(conn._parser, _RESP3Parser):
        conn._parser._buffer._sock = mock_socket
    else:
        conn._parser._sock = mock_socket
    for i in range(100):
        try:
            response = conn.read_response(disconnect_on_error=False)
            break
        except MockSocket.TestError:
            pass

    else:
        pytest.fail("didn't receive a response")
    assert response
    assert i > 0


@pytest.mark.onlynoncluster
@pytest.mark.parametrize(
    "Class",
    [
        Connection,
        SSLConnection,
        UnixDomainSocketConnection,
    ],
)
def test_pack_command(Class):
    """
    This test verifies that the pack_command works
    on all supported connections. #2581
    """
    cmd = (
        "HSET",
        "foo",
        "key",
        "value1",
        b"key_b",
        b"bytes str",
        b"key_i",
        67,
        "key_f",
        3.14159265359,
    )
    expected = (
        b"*10\r\n$4\r\nHSET\r\n$3\r\nfoo\r\n$3\r\nkey\r\n$6\r\nvalue1\r\n"
        b"$5\r\nkey_b\r\n$9\r\nbytes str\r\n$5\r\nkey_i\r\n$2\r\n67\r\n$5"
        b"\r\nkey_f\r\n$13\r\n3.14159265359\r\n"
    )

    actual = Class().pack_command(*cmd)[0]
    assert actual == expected, f"actual = {actual}, expected = {expected}"


@pytest.mark.onlynoncluster
def test_create_single_connection_client_from_url():
    client = valkey.Valkey.from_url(
        "valkey://localhost:6379/0?", single_connection_client=True
    )
    assert client.connection is not None


@pytest.mark.parametrize("from_url", (True, False), ids=("from_url", "from_args"))
def test_pool_auto_close(request, from_url):
    """Verify that basic Valkey instances have auto_close_connection_pool set to True"""

    url: str = request.config.getoption("--valkey-url")
    url_args = parse_url(url)

    def get_valkey_connection():
        if from_url:
            return Valkey.from_url(url)
        return Valkey(**url_args)

    r1 = get_valkey_connection()
    assert r1.auto_close_connection_pool is True
    r1.close()


@pytest.mark.parametrize("from_url", (True, False), ids=("from_url", "from_args"))
def test_valkey_connection_pool(request, from_url):
    """Verify that basic Valkey instances using `connection_pool`
    have auto_close_connection_pool set to False"""

    url: str = request.config.getoption("--valkey-url")
    url_args = parse_url(url)

    pool = None

    def get_valkey_connection():
        nonlocal pool
        if from_url:
            pool = ConnectionPool.from_url(url)
        else:
            pool = ConnectionPool(**url_args)
        return Valkey(connection_pool=pool)

    called = 0

    def mock_disconnect(_):
        nonlocal called
        called += 1

    with patch.object(ConnectionPool, "disconnect", mock_disconnect):
        with get_valkey_connection() as r1:
            assert r1.auto_close_connection_pool is False

    assert called == 0
    pool.disconnect()


@pytest.mark.parametrize("from_url", (True, False), ids=("from_url", "from_args"))
def test_valkey_from_pool(request, from_url):
    """Verify that basic Valkey instances created using `from_pool()`
    have auto_close_connection_pool set to True"""

    url: str = request.config.getoption("--valkey-url")
    url_args = parse_url(url)

    pool = None

    def get_valkey_connection():
        nonlocal pool
        if from_url:
            pool = ConnectionPool.from_url(url)
        else:
            pool = ConnectionPool(**url_args)
        return Valkey.from_pool(pool)

    called = 0

    def mock_disconnect(_):
        nonlocal called
        called += 1

    with patch.object(ConnectionPool, "disconnect", mock_disconnect):
        with get_valkey_connection() as r1:
            assert r1.auto_close_connection_pool is True

    assert called == 1
    pool.disconnect()


@pytest.mark.parametrize(
    "conn, error, expected_message",
    [
        (SSLConnection(), OSError(), "Error connecting to localhost:6379."),
        (SSLConnection(), OSError(12), "Error 12 connecting to localhost:6379."),
        (
            SSLConnection(),
            OSError(12, "Some Error"),
            "Error 12 connecting to localhost:6379. Some Error.",
        ),
        (
            UnixDomainSocketConnection(path="unix:///tmp/valkey.sock"),
            OSError(),
            "Error connecting to unix:///tmp/valkey.sock.",
        ),
        (
            UnixDomainSocketConnection(path="unix:///tmp/valkey.sock"),
            OSError(12),
            "Error 12 connecting to unix:///tmp/valkey.sock.",
        ),
        (
            UnixDomainSocketConnection(path="unix:///tmp/valkey.sock"),
            OSError(12, "Some Error"),
            "Error 12 connecting to unix:///tmp/valkey.sock. Some Error.",
        ),
    ],
)
def test_format_error_message(conn, error, expected_message):
    """Test that the _error_message function formats errors correctly"""
    error_message = conn._error_message(error)
    assert error_message == expected_message


def test_network_connection_failure():
    with pytest.raises(ConnectionError) as e:
        valkey = Valkey(port=9999)
        valkey.set("a", "b")
    assert str(e.value) == "Error 111 connecting to localhost:9999. Connection refused."


def test_unix_socket_connection_failure():
    with pytest.raises(ConnectionError) as e:
        valkey = Valkey(unix_socket_path="unix:///tmp/a.sock")
        valkey.set("a", "b")
    assert (
        str(e.value)
        == "Error 2 connecting to unix:///tmp/a.sock. No such file or directory."
    )


def test_parsing_unix_socket_relative_path():
    parsed = parse_url("unix:./valkey.sock")
    assert parsed["path"] == "./valkey.sock"
    assert parsed["connection_class"] is UnixDomainSocketConnection
    assert len(parsed) == 2



from .base import BaseParser, _AsyncRESPBase
from .commands import AsyncCommandsParser, CommandsParser
from .encoders import Encoder
from .libvalkey import _AsyncLibvalkeyParser, _LibvalkeyParser
from .resp2 import _AsyncRESP2Parser, _RESP2Parser
from .resp3 import _AsyncRESP3Parser, _RESP3Parser
from .url_parser import parse_url

__all__ = [
    "AsyncCommandsParser",
    "_AsyncLibvalkeyParser",
    "_AsyncRESPBase",
    "_AsyncRESP2Parser",
    "_AsyncRESP3Parser",
    "CommandsParser",
    "Encoder",
    "BaseParser",
    "_LibvalkeyParser",
    "_RESP2Parser",
    "_RESP3Parser",
    "parse_url",
]



import time

import pytest
import pytest_asyncio
from valkey._cache import EvictionPolicy, _LocalCache
from valkey.utils import LIBVALKEY_AVAILABLE


@pytest_asyncio.fixture
async def r(request, create_valkey):
    cache = request.param.get("cache")
    kwargs = request.param.get("kwargs", {})
    r = await create_valkey(protocol=3, client_cache=cache, **kwargs)
    yield r, cache


@pytest_asyncio.fixture()
async def local_cache():
    yield _LocalCache()


@pytest.mark.skipif(LIBVALKEY_AVAILABLE, reason="PythonParser only")
class TestLocalCache:
    @pytest.mark.parametrize("r", [{"cache": _LocalCache()}], indirect=True)
    @pytest.mark.onlynoncluster
    async def test_get_from_cache(self, r, r2):
        r, cache = r
        # add key to valkey
        await r.set("foo", "bar")
        # get key from valkey and save in local cache
        assert await r.get("foo") == b"bar"
        # get key from local cache
        assert cache.get(("GET", "foo")) == b"bar"
        # change key in valkey (cause invalidation)
        await r2.set("foo", "barbar")
        # send any command to valkey (process invalidation in background)
        await r.ping()
        # the command is not in the local cache anymore
        assert cache.get(("GET", "foo")) is None
        # get key from valkey
        assert await r.get("foo") == b"barbar"

    @pytest.mark.parametrize("r", [{"cache": _LocalCache(max_size=3)}], indirect=True)
    async def test_cache_lru_eviction(self, r):
        r, cache = r
        # add 3 keys to valkey
        await r.set("foo", "bar")
        await r.set("foo2", "bar2")
        await r.set("foo3", "bar3")
        # get 3 keys from valkey and save in local cache
        assert await r.get("foo") == b"bar"
        assert await r.get("foo2") == b"bar2"
        assert await r.get("foo3") == b"bar3"
        # get the 3 keys from local cache
        assert cache.get(("GET", "foo")) == b"bar"
        assert cache.get(("GET", "foo2")) == b"bar2"
        assert cache.get(("GET", "foo3")) == b"bar3"
        # add 1 more key to valkey (exceed the max size)
        await r.set("foo4", "bar4")
        assert await r.get("foo4") == b"bar4"
        # the first key is not in the local cache anymore
        assert cache.get(("GET", "foo")) is None

    @pytest.mark.parametrize("r", [{"cache": _LocalCache(ttl=1)}], indirect=True)
    async def test_cache_ttl(self, r):
        r, cache = r
        # add key to valkey
        await r.set("foo", "bar")
        # get key from valkey and save in local cache
        assert await r.get("foo") == b"bar"
        # get key from local cache
        assert cache.get(("GET", "foo")) == b"bar"
        # wait for the key to expire
        time.sleep(1)
        # the key is not in the local cache anymore
        assert cache.get(("GET", "foo")) is None

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(max_size=3, eviction_policy=EvictionPolicy.LFU)}],
        indirect=True,
    )
    async def test_cache_lfu_eviction(self, r):
        r, cache = r
        # add 3 keys to valkey
        await r.set("foo", "bar")
        await r.set("foo2", "bar2")
        await r.set("foo3", "bar3")
        # get 3 keys from valkey and save in local cache
        assert await r.get("foo") == b"bar"
        assert await r.get("foo2") == b"bar2"
        assert await r.get("foo3") == b"bar3"
        # change the order of the keys in the cache
        assert cache.get(("GET", "foo")) == b"bar"
        assert cache.get(("GET", "foo")) == b"bar"
        assert cache.get(("GET", "foo3")) == b"bar3"
        # add 1 more key to valkey (exceed the max size)
        await r.set("foo4", "bar4")
        assert await r.get("foo4") == b"bar4"
        # test the eviction policy
        assert len(cache.cache) == 3
        assert cache.get(("GET", "foo")) == b"bar"
        assert cache.get(("GET", "foo2")) is None

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    @pytest.mark.onlynoncluster
    async def test_cache_decode_response(self, r):
        r, cache = r
        await r.set("foo", "bar")
        # get key from valkey and save in local cache
        assert await r.get("foo") == "bar"
        # get key from local cache
        assert cache.get(("GET", "foo")) == "bar"
        # change key in valkey (cause invalidation)
        await r.set("foo", "barbar")
        # send any command to valkey (process invalidation in background)
        await r.ping()
        # the command is not in the local cache anymore
        assert cache.get(("GET", "foo")) is None
        # get key from valkey
        assert await r.get("foo") == "barbar"

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"cache_deny_list": ["LLEN"]}}],
        indirect=True,
    )
    async def test_cache_deny_list(self, r):
        r, cache = r
        # add list to valkey
        await r.lpush("mylist", "foo", "bar", "baz")
        assert await r.llen("mylist") == 3
        assert await r.lindex("mylist", 1) == b"bar"
        assert cache.get(("LLEN", "mylist")) is None
        assert cache.get(("LINDEX", "mylist", 1)) == b"bar"

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"cache_allow_list": ["LLEN"]}}],
        indirect=True,
    )
    async def test_cache_allow_list(self, r):
        r, cache = r
        # add list to valkey
        await r.lpush("mylist", "foo", "bar", "baz")
        assert await r.llen("mylist") == 3
        assert await r.lindex("mylist", 1) == b"bar"
        assert cache.get(("LLEN", "mylist")) == 3
        assert cache.get(("LINDEX", "mylist", 1)) is None

    @pytest.mark.parametrize("r", [{"cache": _LocalCache()}], indirect=True)
    async def test_cache_return_copy(self, r):
        r, cache = r
        await r.lpush("mylist", "foo", "bar", "baz")
        assert await r.lrange("mylist", 0, -1) == [b"baz", b"bar", b"foo"]
        res = cache.get(("LRANGE", "mylist", 0, -1))
        assert res == [b"baz", b"bar", b"foo"]
        res.append(b"new")
        check = cache.get(("LRANGE", "mylist", 0, -1))
        assert check == [b"baz", b"bar", b"foo"]

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    @pytest.mark.onlynoncluster
    async def test_csc_not_cause_disconnects(self, r):
        r, cache = r
        id1 = await r.client_id()
        await r.mset({"a": 1, "b": 1, "c": 1, "d": 1, "e": 1})
        assert await r.mget("a", "b", "c", "d", "e") == ["1", "1", "1", "1", "1"]
        id2 = await r.client_id()

        # client should get value from client cache
        assert await r.mget("a", "b", "c", "d", "e") == ["1", "1", "1", "1", "1"]
        assert cache.get(("MGET", "a", "b", "c", "d", "e")) == [
            "1",
            "1",
            "1",
            "1",
            "1",
        ]

        await r.mset({"a": 2, "b": 2, "c": 2, "d": 2, "e": 2})
        id3 = await r.client_id()
        # client should get value from valkey server post invalidate messages
        assert await r.mget("a", "b", "c", "d", "e") == ["2", "2", "2", "2", "2"]

        await r.mset({"a": 3, "b": 3, "c": 3, "d": 3, "e": 3})
        # need to check that we get correct value 3 and not 2
        assert await r.mget("a", "b", "c", "d", "e") == ["3", "3", "3", "3", "3"]
        # client should get value from client cache
        assert await r.mget("a", "b", "c", "d", "e") == ["3", "3", "3", "3", "3"]

        await r.mset({"a": 4, "b": 4, "c": 4, "d": 4, "e": 4})
        # need to check that we get correct value 4 and not 3
        assert await r.mget("a", "b", "c", "d", "e") == ["4", "4", "4", "4", "4"]
        # client should get value from client cache
        assert await r.mget("a", "b", "c", "d", "e") == ["4", "4", "4", "4", "4"]
        id4 = await r.client_id()
        assert id1 == id2 == id3 == id4

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    async def test_execute_command_keys_provided(self, r):
        r, cache = r
        assert await r.execute_command("SET", "b", "2") is True
        assert await r.execute_command("GET", "b", keys=["b"]) == "2"
        assert cache.get(("GET", "b")) == "2"

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    async def test_execute_command_keys_not_provided(self, r):
        r, cache = r
        assert await r.execute_command("SET", "b", "2") is True
        assert (
            await r.execute_command("GET", "b") == "2"
        )  # keys not provided, not cached
        assert cache.get(("GET", "b")) is None

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    async def test_delete_one_command(self, r):
        r, cache = r
        assert await r.mset({"a{a}": 1, "b{a}": 1}) is True
        assert await r.set("c", 1) is True
        assert await r.mget("a{a}", "b{a}") == ["1", "1"]
        assert await r.get("c") == "1"
        # values should be in local cache
        assert cache.get(("MGET", "a{a}", "b{a}")) == ["1", "1"]
        assert cache.get(("GET", "c")) == "1"
        # delete one command from the cache
        r.delete_command_from_cache(("MGET", "a{a}", "b{a}"))
        # the other command is still in the local cache anymore
        assert cache.get(("MGET", "a{a}", "b{a}")) is None
        assert cache.get(("GET", "c")) == "1"
        # get from valkey
        assert await r.mget("a{a}", "b{a}") == ["1", "1"]
        assert await r.get("c") == "1"

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    async def test_invalidate_key(self, r):
        r, cache = r
        assert await r.mset({"a{a}": 1, "b{a}": 1}) is True
        assert await r.set("c", 1) is True
        assert await r.mget("a{a}", "b{a}") == ["1", "1"]
        assert await r.get("c") == "1"
        # values should be in local cache
        assert cache.get(("MGET", "a{a}", "b{a}")) == ["1", "1"]
        assert cache.get(("GET", "c")) == "1"
        # invalidate one key from the cache
        r.invalidate_key_from_cache("b{a}")
        # one other command is still in the local cache anymore
        assert cache.get(("MGET", "a{a}", "b{a}")) is None
        assert cache.get(("GET", "c")) == "1"
        # get from valkey
        assert await r.mget("a{a}", "b{a}") == ["1", "1"]
        assert await r.get("c") == "1"

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    async def test_flush_entire_cache(self, r):
        r, cache = r
        assert await r.mset({"a{a}": 1, "b{a}": 1}) is True
        assert await r.set("c", 1) is True
        assert await r.mget("a{a}", "b{a}") == ["1", "1"]
        assert await r.get("c") == "1"
        # values should be in local cache
        assert cache.get(("MGET", "a{a}", "b{a}")) == ["1", "1"]
        assert cache.get(("GET", "c")) == "1"
        # flush the local cache
        r.flush_cache()
        # the commands are not in the local cache anymore
        assert cache.get(("MGET", "a{a}", "b{a}")) is None
        assert cache.get(("GET", "c")) is None
        # get from valkey
        assert await r.mget("a{a}", "b{a}") == ["1", "1"]
        assert await r.get("c") == "1"


@pytest.mark.skipif(LIBVALKEY_AVAILABLE, reason="PythonParser only")
@pytest.mark.onlycluster
class TestClusterLocalCache:
    @pytest.mark.parametrize("r", [{"cache": _LocalCache()}], indirect=True)
    async def test_get_from_cache(self, r, r2):
        r, cache = r
        # add key to valkey
        await r.set("foo", "bar")
        # get key from valkey and save in local cache
        assert await r.get("foo") == b"bar"
        # get key from local cache
        assert cache.get(("GET", "foo")) == b"bar"
        # change key in valkey (cause invalidation)
        await r2.set("foo", "barbar")
        # send any command to valkey (process invalidation in background)
        node = r.get_node_from_key("foo")
        await r.ping(target_nodes=node)
        # the command is not in the local cache anymore
        assert cache.get(("GET", "foo")) is None
        # get key from valkey
        assert await r.get("foo") == b"barbar"

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    async def test_cache_decode_response(self, r):
        r, cache = r
        await r.set("foo", "bar")
        # get key from valkey and save in local cache
        assert await r.get("foo") == "bar"
        # get key from local cache
        assert cache.get(("GET", "foo")) == "bar"
        # change key in valkey (cause invalidation)
        await r.set("foo", "barbar")
        # send any command to valkey (process invalidation in background)
        node = r.get_node_from_key("foo")
        await r.ping(target_nodes=node)
        # the command is not in the local cache anymore
        assert cache.get(("GET", "foo")) is None
        # get key from valkey
        assert await r.get("foo") == "barbar"

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    async def test_execute_command_keys_provided(self, r):
        r, cache = r
        assert await r.execute_command("SET", "b", "2") is True
        assert await r.execute_command("GET", "b", keys=["b"]) == "2"
        assert cache.get(("GET", "b")) == "2"

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    async def test_execute_command_keys_not_provided(self, r):
        r, cache = r
        assert await r.execute_command("SET", "b", "2") is True
        assert (
            await r.execute_command("GET", "b") == "2"
        )  # keys not provided, not cached
        assert cache.get(("GET", "b")) is None


@pytest.mark.skipif(LIBVALKEY_AVAILABLE, reason="PythonParser only")
@pytest.mark.onlynoncluster
class TestSentinelLocalCache:

    async def test_get_from_cache(self, local_cache, master):
        await master.set("foo", "bar")
        # get key from valkey and save in local cache
        assert await master.get("foo") == b"bar"
        # get key from local cache
        assert local_cache.get(("GET", "foo")) == b"bar"
        # change key in valkey (cause invalidation)
        await master.set("foo", "barbar")
        # send any command to valkey (process invalidation in background)
        await master.ping()
        # the command is not in the local cache anymore
        assert local_cache.get(("GET", "foo")) is None
        # get key from valkey
        assert await master.get("foo") == b"barbar"

    @pytest.mark.parametrize(
        "sentinel_setup",
        [{"kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    async def test_cache_decode_response(self, local_cache, sentinel_setup, master):
        await master.set("foo", "bar")
        # get key from valkey and save in local cache
        assert await master.get("foo") == "bar"
        # get key from local cache
        assert local_cache.get(("GET", "foo")) == "bar"
        # change key in valkey (cause invalidation)
        await master.set("foo", "barbar")
        # send any command to valkey (process invalidation in background)
        await master.ping()
        # the command is not in the local cache anymore
        assert local_cache.get(("GET", "foo")) is None
        # get key from valkey
        assert await master.get("foo") == "barbar"



import argparse
import math
import time
from collections.abc import Callable
from typing import TypeVar
from unittest import mock
from unittest.mock import Mock
from urllib.parse import urlparse

import pytest
import valkey
from packaging.version import Version
from valkey import Sentinel
from valkey.backoff import NoBackoff
from valkey.connection import Connection, parse_url
from valkey.exceptions import ValkeyClusterException
from valkey.retry import Retry

VALKEY_INFO = {}
default_valkey_url = "valkey://localhost:6379/0"
default_protocol = "2"
default_valkeymod_url = "valkey://localhost:6379"

# default ssl client ignores verification for the purpose of testing
default_valkey_ssl_url = "valkeys://localhost:6666"
default_cluster_nodes = 6

_DecoratedTest = TypeVar("_DecoratedTest", bound="Callable")
_TestDecorator = Callable[[_DecoratedTest], _DecoratedTest]


def pytest_addoption(parser):
    parser.addoption(
        "--valkey-url",
        default=default_valkey_url,
        action="store",
        help="Valkey connection string, defaults to `%(default)s`",
    )

    parser.addoption(
        "--protocol",
        default=default_protocol,
        action="store",
        help="Protocol version, defaults to `%(default)s`",
    )
    parser.addoption(
        "--valkey-ssl-url",
        default=default_valkey_ssl_url,
        action="store",
        help="Valkey SSL connection string, defaults to `%(default)s`",
    )

    parser.addoption(
        "--valkey-cluster-nodes",
        default=default_cluster_nodes,
        action="store",
        help="The number of cluster nodes that need to be "
        "available before the test can start,"
        " defaults to `%(default)s`",
    )

    parser.addoption(
        "--uvloop", action=argparse.BooleanOptionalAction, help="Run tests with uvloop"
    )

    parser.addoption(
        "--sentinels",
        action="store",
        default="localhost:26379,localhost:26380,localhost:26381",
        help="Comma-separated list of sentinel IPs and ports",
    )
    parser.addoption(
        "--master-service",
        action="store",
        default="valkey-py-test",
        help="Name of the Valkey master service that the sentinels are monitoring",
    )


def _get_info(valkey_url):
    client = valkey.Valkey.from_url(valkey_url)
    info = client.info()
    client.connection_pool.disconnect()
    return info


def pytest_sessionstart(session):
    # during test discovery, e.g. with VS Code, we may not
    # have a server running.
    valkey_url = session.config.getoption("--valkey-url")
    try:
        info = _get_info(valkey_url)
        version = info["valkey_version"]
        arch_bits = info["arch_bits"]
        cluster_enabled = info["cluster_enabled"]
    except valkey.ConnectionError:
        # provide optimistic defaults
        info = {}
        version = "10.0.0"
        arch_bits = 64
        cluster_enabled = False
    VALKEY_INFO["version"] = version
    VALKEY_INFO["arch_bits"] = arch_bits
    VALKEY_INFO["cluster_enabled"] = cluster_enabled
    # store VALKEY_INFO in config so that it is available from "condition strings"
    session.config.VALKEY_INFO = VALKEY_INFO

    # module info
    try:
        VALKEY_INFO["modules"] = info["modules"]
    except (KeyError, valkey.exceptions.ConnectionError):
        pass

    if cluster_enabled:
        cluster_nodes = session.config.getoption("--valkey-cluster-nodes")
        wait_for_cluster_creation(valkey_url, cluster_nodes)

    use_uvloop = session.config.getoption("--uvloop")

    if use_uvloop:
        try:
            import uvloop

            uvloop.install()
        except ImportError as e:
            raise RuntimeError("Cannot import uvloop, make sure it is installed") from e


def wait_for_cluster_creation(valkey_url, cluster_nodes, timeout=60):
    """Waits for the cluster creation to complete.
    As soon as all :cluster_nodes: nodes become available, the cluster will be
    considered ready.
    :param valkey_url: the cluster's url, e.g. valkey://localhost:16379/0
    :param cluster_nodes: The number of nodes in the cluster
    :param timeout: the amount of time to wait (in seconds).
    """
    now = time.time()
    end_time = now + timeout
    client = None
    print(f"Waiting for {cluster_nodes} cluster nodes to become available")
    while now < end_time:
        try:
            client = valkey.ValkeyCluster.from_url(valkey_url)
            if len(client.get_nodes()) == int(cluster_nodes):
                print("All nodes are available!")
                break
        except ValkeyClusterException:
            pass
        time.sleep(1)
        now = time.time()
    if now >= end_time:
        available_nodes = 0 if client is None else len(client.get_nodes())
        raise ValkeyClusterException(
            f"The cluster did not become available after {timeout} seconds. "
            f"Only {available_nodes} nodes out of {cluster_nodes} are available"
        )


def skip_if_server_version_lt(min_version: str) -> _TestDecorator:
    valkey_version = VALKEY_INFO.get("version", "0")
    check = Version(valkey_version) < Version(min_version)
    return pytest.mark.skipif(check, reason=f"Valkey version required >= {min_version}")


def skip_if_server_version_gte(min_version: str) -> _TestDecorator:
    valkey_version = VALKEY_INFO.get("version", "0")
    check = Version(valkey_version) >= Version(min_version)
    return pytest.mark.skipif(check, reason=f"Valkey version required < {min_version}")


def skip_if_version_is_one_of(versions: list[str]) -> _TestDecorator:
    valkey_version = VALKEY_INFO.get("version", "0")
    check = Version(valkey_version) in [Version(v) for v in versions]
    return pytest.mark.skipif(
        check, reason=f"Valkey version required not in {versions}"
    )


def skip_unless_arch_bits(arch_bits: int) -> _TestDecorator:
    return pytest.mark.skipif(
        VALKEY_INFO.get("arch_bits", "") != arch_bits,
        reason=f"server is not {arch_bits}-bit",
    )


def skip_ifmodversion_lt(min_version: str, module_name: str):
    try:
        modules = VALKEY_INFO["modules"]
    except KeyError:
        return pytest.mark.skipif(True, reason="Valkey server does not have modules")
    if modules == []:
        return pytest.mark.skipif(True, reason="No valkey modules found")

    for j in modules:
        if module_name == j.get("name"):
            version = j.get("ver")
            mv = int(
                "".join([f"{int(segment):02}" for segment in min_version.split(".")])
            )
            check = version < mv
            return pytest.mark.skipif(check, reason="Valkey module version")

    raise AttributeError(f"No valkey module named {module_name}")


def skip_if_nocryptography() -> _TestDecorator:
    try:
        import cryptography  # noqa

        return pytest.mark.skipif(False, reason="Cryptography dependency found")
    except ImportError:
        return pytest.mark.skipif(True, reason="No cryptography dependency")


def skip_if_cryptography() -> _TestDecorator:
    try:
        import cryptography  # noqa

        return pytest.mark.skipif(True, reason="Cryptography dependency found")
    except ImportError:
        return pytest.mark.skipif(False, reason="No cryptography dependency")


def _get_client(
    cls, request, single_connection_client=True, flushdb=True, from_url=None, **kwargs
):
    """Helper for fixtures or tests that need a Valkey client.

    Uses the "--valkey-url" command line argument for connection info. Unlike
    ConnectionPool.from_url, keyword arguments to this function override
    values specified in the URL.
    """
    if from_url is None:
        valkey_url = request.config.getoption("--valkey-url")
    else:
        valkey_url = from_url
    if "protocol" not in valkey_url and kwargs.get("protocol") is None:
        kwargs["protocol"] = request.config.getoption("--protocol")

    cluster_mode = VALKEY_INFO["cluster_enabled"]
    if not cluster_mode:
        url_options = parse_url(valkey_url)
        url_options.update(kwargs)
        pool = valkey.ConnectionPool(**url_options)
        client = cls(connection_pool=pool)
    else:
        client = valkey.ValkeyCluster.from_url(valkey_url, **kwargs)
        single_connection_client = False
    if single_connection_client:
        client = client.client()
    if request:

        def teardown():
            if not cluster_mode:
                if flushdb:
                    try:
                        client.flushdb()
                    except valkey.ConnectionError:
                        # handle cases where a test disconnected a client
                        # just manually retry the flushdb
                        client.flushdb()
                client.close()
                client.connection_pool.disconnect()
            else:
                cluster_teardown(client, flushdb)

        request.addfinalizer(teardown)
    return client


def cluster_teardown(client, flushdb):
    if flushdb:
        try:
            client.flushdb(target_nodes="primaries")
        except valkey.ConnectionError:
            # handle cases where a test disconnected a client
            # just manually retry the flushdb
            client.flushdb(target_nodes="primaries")
    client.close()
    client.disconnect_connection_pools()


@pytest.fixture
def r(request):
    with _get_client(valkey.Valkey, request) as client:
        yield client


@pytest.fixture
def decoded_r(request):
    with _get_client(valkey.Valkey, request, decode_responses=True) as client:
        yield client


@pytest.fixture
def r_timeout(request):
    with _get_client(valkey.Valkey, request, socket_timeout=1) as client:
        yield client


@pytest.fixture
def r2(request):
    """A second client for tests that need multiple."""
    with _get_client(valkey.Valkey, request) as client:
        yield client


@pytest.fixture
def sslclient(request):
    with _get_client(valkey.Valkey, request, ssl=True) as client:
        yield client


@pytest.fixture
def sentinel_setup(local_cache, request):
    sentinel_ips = request.config.getoption("--sentinels")
    sentinel_endpoints = [
        (ip.strip(), int(port.strip()))
        for ip, port in (endpoint.split(":") for endpoint in sentinel_ips.split(","))
    ]
    kwargs = request.param.get("kwargs", {}) if hasattr(request, "param") else {}
    sentinel = Sentinel(
        sentinel_endpoints,
        socket_timeout=0.1,
        client_cache=local_cache,
        protocol=3,
        **kwargs,
    )
    yield sentinel
    for s in sentinel.sentinels:
        s.close()


@pytest.fixture
def master(request, sentinel_setup):
    master_service = request.config.getoption("--master-service")
    master = sentinel_setup.master_for(master_service)
    yield master
    master.close()


def _gen_cluster_mock_resp(r, response):
    connection = Mock(spec=Connection)
    connection.retry = Retry(NoBackoff(), 0)
    connection.read_response.return_value = response
    connection._get_from_local_cache.return_value = None
    with mock.patch.object(r, "connection", connection):
        yield r


@pytest.fixture
def mock_cluster_resp_ok(request, **kwargs):
    r = _get_client(valkey.Valkey, request, **kwargs)
    yield from _gen_cluster_mock_resp(r, "OK")


@pytest.fixture
def mock_cluster_resp_int(request, **kwargs):
    r = _get_client(valkey.Valkey, request, **kwargs)
    yield from _gen_cluster_mock_resp(r, 2)


@pytest.fixture
def mock_cluster_resp_info(request, **kwargs):
    r = _get_client(valkey.Valkey, request, **kwargs)
    response = (
        "cluster_state:ok\r\ncluster_slots_assigned:16384\r\n"
        "cluster_slots_ok:16384\r\ncluster_slots_pfail:0\r\n"
        "cluster_slots_fail:0\r\ncluster_known_nodes:7\r\n"
        "cluster_size:3\r\ncluster_current_epoch:7\r\n"
        "cluster_my_epoch:2\r\ncluster_stats_messages_sent:170262\r\n"
        "cluster_stats_messages_received:105653\r\n"
    )
    yield from _gen_cluster_mock_resp(r, response)


@pytest.fixture
def mock_cluster_resp_nodes(request, **kwargs):
    r = _get_client(valkey.Valkey, request, **kwargs)
    response = (
        "c8253bae761cb1ecb2b61857d85dfe455a0fec8b 172.17.0.7:7006 "
        "slave aa90da731f673a99617dfe930306549a09f83a6b 0 "
        "1447836263059 5 connected\n"
        "9bd595fe4821a0e8d6b99d70faa660638a7612b3 172.17.0.7:7008 "
        "master - 0 1447836264065 0 connected\n"
        "aa90da731f673a99617dfe930306549a09f83a6b 172.17.0.7:7003 "
        "myself,master - 0 0 2 connected 5461-10922\n"
        "1df047e5a594f945d82fc140be97a1452bcbf93e 172.17.0.7:7007 "
        "slave 19efe5a631f3296fdf21a5441680f893e8cc96ec 0 "
        "1447836262556 3 connected\n"
        "4ad9a12e63e8f0207025eeba2354bcf4c85e5b22 172.17.0.7:7005 "
        "master - 0 1447836262555 7 connected 0-5460\n"
        "19efe5a631f3296fdf21a5441680f893e8cc96ec 172.17.0.7:7004 "
        "master - 0 1447836263562 3 connected 10923-16383\n"
        "fbb23ed8cfa23f17eaf27ff7d0c410492a1093d6 172.17.0.7:7002 "
        "master,fail - 1447829446956 1447829444948 1 disconnected\n"
    )
    yield from _gen_cluster_mock_resp(r, response)


@pytest.fixture
def mock_cluster_resp_slaves(request, **kwargs):
    r = _get_client(valkey.Valkey, request, **kwargs)
    response = (
        "['1df047e5a594f945d82fc140be97a1452bcbf93e 172.17.0.7:7007 "
        "slave 19efe5a631f3296fdf21a5441680f893e8cc96ec 0 "
        "1447836789290 3 connected']"
    )
    yield from _gen_cluster_mock_resp(r, response)


@pytest.fixture(scope="session")
def master_host(request):
    url = request.config.getoption("--valkey-url")
    parts = urlparse(url)
    return parts.hostname, (parts.port or 6379)


@pytest.fixture
def valkey_version():
    return Version(VALKEY_INFO["version"])


def wait_for_command(client, monitor, command, key=None):
    # issue a command with a key name that's local to this process.
    # if we find a command with our key before the command we're waiting
    # for, something went wrong
    if key is None:
        # generate key
        id_str = str(client.client_id())
        key = f"__VALKEY-PY-{id_str}__"
    client.get(key)
    while True:
        monitor_response = monitor.next_command()
        if command in monitor_response["command"]:
            return monitor_response
        if key in monitor_response["command"]:
            return None


def is_resp2_connection(r):
    if isinstance(r, valkey.Valkey) or isinstance(r, valkey.asyncio.Valkey):
        protocol = r.connection_pool.connection_kwargs.get("protocol")
    elif isinstance(r, valkey.cluster.AbstractValkeyCluster):
        protocol = r.nodes_manager.connection_kwargs.get("protocol")
    return protocol in ["2", 2, None]


def get_protocol_version(r):
    if isinstance(r, valkey.Valkey) or isinstance(r, valkey.asyncio.Valkey):
        return r.connection_pool.connection_kwargs.get("protocol")
    elif isinstance(r, valkey.cluster.AbstractValkeyCluster):
        return r.nodes_manager.connection_kwargs.get("protocol")


def assert_resp_response(r, response, resp2_expected, resp3_expected):
    protocol = get_protocol_version(r)
    if protocol in [2, "2", None]:
        assert response == resp2_expected
    else:
        assert response == resp3_expected


def assert_geo_is_close(coords, expected_coords):
    """Verifies that the coordinates are close within the floating point tolerance.

    Valkey uses 52-bit precision
    """
    for a, b in zip(coords, expected_coords):
        assert math.isclose(a, b)


def assert_resp_response_isclose(r, response, resp2_expected, resp3_expected):
    """Verifies that the responses are close within the floating point tolerance."""
    protocol = get_protocol_version(r)

    if protocol in [2, "2", None]:
        assert_geo_is_close(response[0], resp2_expected[0])
        assert response[1:] == resp2_expected[1:]
    else:
        assert_geo_is_close(response[0], resp3_expected[0])
        assert response[1:] == resp3_expected[1:]


def assert_resp_response_in(r, response, resp2_expected, resp3_expected):
    protocol = get_protocol_version(r)
    if protocol in [2, "2", None]:
        assert response in resp2_expected
    else:
        assert response in resp3_expected



import re
from types import MappingProxyType
from typing import Callable, Mapping, Optional
from urllib.parse import ParseResult, parse_qs, unquote, urlparse

from valkey.asyncio.connection import (
    ConnectKwargs,
)
from valkey.asyncio.connection import SSLConnection as SSLConnectionAsync
from valkey.asyncio.connection import (
    UnixDomainSocketConnection as UnixDomainSocketConnectionAsync,
)
from valkey.connection import SSLConnection, UnixDomainSocketConnection


def to_bool(value) -> Optional[bool]:
    if value is None or value == "":
        return None
    if isinstance(value, str) and value.upper() in FALSE_STRINGS:
        return False
    return bool(value)


FALSE_STRINGS = ("0", "F", "FALSE", "N", "NO")

URL_QUERY_ARGUMENT_PARSERS: Mapping[str, Callable[..., object]] = MappingProxyType(
    {
        "db": int,
        "socket_timeout": float,
        "socket_connect_timeout": float,
        "socket_keepalive": to_bool,
        "retry_on_timeout": to_bool,
        "max_connections": int,
        "health_check_interval": int,
        "ssl_check_hostname": to_bool,
        "timeout": float,
    }
)


def parse_url(url: str, async_connection: bool):
    supported_schemes = ["valkey", "valkeys", "redis", "rediss", "unix"]
    parsed: ParseResult = urlparse(url)
    kwargs: ConnectKwargs = {}
    lower_url = url.lower()
    pattern = re.compile(
        r"^(?:" + "|".join(map(re.escape, supported_schemes)) + r")://"
    )
    if not pattern.match(lower_url) and not lower_url.startswith("unix:"):
        raise ValueError(
            f"Valkey URL must specify one of the following schemes {supported_schemes}"
        )

    for name, value_list in parse_qs(parsed.query).items():
        if value_list and len(value_list) > 0:
            value = unquote(value_list[0])
            parser = URL_QUERY_ARGUMENT_PARSERS.get(name)
            if parser:
                try:
                    kwargs[name] = parser(value)
                except (TypeError, ValueError):
                    raise ValueError(f"Invalid value for `{name}` in connection URL.")
            else:
                kwargs[name] = value

    if parsed.username:
        kwargs["username"] = unquote(parsed.username)
    if parsed.password:
        kwargs["password"] = unquote(parsed.password)

    # We only support valkey://, valkeys://, redis://, rediss://, and unix:// schemes.
    if parsed.scheme == "unix":
        if parsed.path:
            kwargs["path"] = unquote(parsed.path)
        kwargs["connection_class"] = (
            UnixDomainSocketConnectionAsync
            if async_connection
            else UnixDomainSocketConnection
        )

    elif parsed.scheme in supported_schemes:
        if parsed.hostname:
            kwargs["host"] = unquote(parsed.hostname)
        if parsed.port:
            kwargs["port"] = int(parsed.port)

        # If there's a path argument, use it as the db argument if a
        # querystring value wasn't specified
        if parsed.path and "db" not in kwargs:
            try:
                kwargs["db"] = int(unquote(parsed.path).replace("/", ""))
            except (AttributeError, ValueError):
                pass

        if parsed.scheme in ("valkeys", "rediss"):
            kwargs["connection_class"] = (
                SSLConnectionAsync if async_connection else SSLConnection
            )
    else:
        raise ValueError(
            f"Valkey URL must specify one of the following schemes {supported_schemes}"
        )

    return kwargs



import copy
import random
import time
from abc import ABC, abstractmethod
from collections import OrderedDict, defaultdict
from enum import Enum
from typing import List, Sequence, Union

from valkey.typing import KeyT, ResponseT


class EvictionPolicy(str, Enum):
    LRU = "lru"
    LFU = "lfu"
    RANDOM = "random"


DEFAULT_EVICTION_POLICY = EvictionPolicy.LRU

DEFAULT_DENY_LIST = [
    "BF.CARD",
    "BF.DEBUG",
    "BF.EXISTS",
    "BF.INFO",
    "BF.MEXISTS",
    "BF.SCANDUMP",
    "CF.COMPACT",
    "CF.COUNT",
    "CF.DEBUG",
    "CF.EXISTS",
    "CF.INFO",
    "CF.MEXISTS",
    "CF.SCANDUMP",
    "CMS.INFO",
    "CMS.QUERY",
    "DUMP",
    "EXPIRETIME",
    "FT.AGGREGATE",
    "FT.ALIASADD",
    "FT.ALIASDEL",
    "FT.ALIASUPDATE",
    "FT.CURSOR",
    "FT.EXPLAIN",
    "FT.EXPLAINCLI",
    "FT.GET",
    "FT.INFO",
    "FT.MGET",
    "FT.PROFILE",
    "FT.SEARCH",
    "FT.SPELLCHECK",
    "FT.SUGGET",
    "FT.SUGLEN",
    "FT.SYNDUMP",
    "FT.TAGVALS",
    "FT._ALIASADDIFNX",
    "FT._ALIASDELIFX",
    "HRANDFIELD",
    "JSON.DEBUG",
    "PEXPIRETIME",
    "PFCOUNT",
    "PTTL",
    "SRANDMEMBER",
    "TDIGEST.BYRANK",
    "TDIGEST.BYREVRANK",
    "TDIGEST.CDF",
    "TDIGEST.INFO",
    "TDIGEST.MAX",
    "TDIGEST.MIN",
    "TDIGEST.QUANTILE",
    "TDIGEST.RANK",
    "TDIGEST.REVRANK",
    "TDIGEST.TRIMMED_MEAN",
    "TOPK.INFO",
    "TOPK.LIST",
    "TOPK.QUERY",
    "TOUCH",
    "TTL",
]

DEFAULT_ALLOW_LIST = [
    "BITCOUNT",
    "BITFIELD_RO",
    "BITPOS",
    "EXISTS",
    "GEODIST",
    "GEOHASH",
    "GEOPOS",
    "GEORADIUSBYMEMBER_RO",
    "GEORADIUS_RO",
    "GEOSEARCH",
    "GET",
    "GETBIT",
    "GETRANGE",
    "HEXISTS",
    "HGET",
    "HGETALL",
    "HKEYS",
    "HLEN",
    "HMGET",
    "HSTRLEN",
    "HVALS",
    "JSON.ARRINDEX",
    "JSON.ARRLEN",
    "JSON.GET",
    "JSON.MGET",
    "JSON.OBJKEYS",
    "JSON.OBJLEN",
    "JSON.RESP",
    "JSON.STRLEN",
    "JSON.TYPE",
    "LCS",
    "LINDEX",
    "LLEN",
    "LPOS",
    "LRANGE",
    "MGET",
    "SCARD",
    "SDIFF",
    "SINTER",
    "SINTERCARD",
    "SISMEMBER",
    "SMEMBERS",
    "SMISMEMBER",
    "SORT_RO",
    "STRLEN",
    "SUBSTR",
    "SUNION",
    "TS.GET",
    "TS.INFO",
    "TS.RANGE",
    "TS.REVRANGE",
    "TYPE",
    "XLEN",
    "XPENDING",
    "XRANGE",
    "XREAD",
    "XREVRANGE",
    "ZCARD",
    "ZCOUNT",
    "ZDIFF",
    "ZINTER",
    "ZINTERCARD",
    "ZLEXCOUNT",
    "ZMSCORE",
    "ZRANGE",
    "ZRANGEBYLEX",
    "ZRANGEBYSCORE",
    "ZRANK",
    "ZREVRANGE",
    "ZREVRANGEBYLEX",
    "ZREVRANGEBYSCORE",
    "ZREVRANK",
    "ZSCORE",
    "ZUNION",
]

_RESPONSE = "response"
_KEYS = "keys"
_CTIME = "ctime"
_ACCESS_COUNT = "access_count"


class AbstractCache(ABC):
    """
    An abstract base class for client caching implementations.
    If you want to implement your own cache you must support these methods.
    """

    @abstractmethod
    def set(
        self,
        command: Union[str, Sequence[str]],
        response: ResponseT,
        keys_in_command: List[KeyT],
    ):
        pass

    @abstractmethod
    def get(self, command: Union[str, Sequence[str]]) -> ResponseT:
        pass

    @abstractmethod
    def delete_command(self, command: Union[str, Sequence[str]]):
        pass

    @abstractmethod
    def delete_commands(self, commands: List[Union[str, Sequence[str]]]):
        pass

    @abstractmethod
    def flush(self):
        pass

    @abstractmethod
    def invalidate_key(self, key: KeyT):
        pass


class _LocalCache(AbstractCache):
    """
    A caching mechanism for storing valkey commands and their responses.

    Args:
        max_size (int): The maximum number of commands to be stored in the cache.
        ttl (int): The time-to-live for each command in seconds.
        eviction_policy (EvictionPolicy): The eviction policy to use for removing commands when the cache is full.

    Attributes:
        max_size (int): The maximum number of commands to be stored in the cache.
        ttl (int): The time-to-live for each command in seconds.
        eviction_policy (EvictionPolicy): The eviction policy used for cache management.
        cache (OrderedDict): The ordered dictionary to store commands and their metadata.
        key_commands_map (defaultdict): A mapping of keys to the set of commands that use each key.
        commands_ttl_list (list): A list to keep track of the commands in the order they were added.  # noqa
    """

    def __init__(
        self,
        max_size: int = 10000,
        ttl: int = 0,
        eviction_policy: EvictionPolicy = DEFAULT_EVICTION_POLICY,
    ):
        self.max_size = max_size
        self.ttl = ttl
        self.eviction_policy = eviction_policy
        self.cache = OrderedDict()
        self.key_commands_map = defaultdict(set)
        self.commands_ttl_list = []

    def set(
        self,
        command: Union[str, Sequence[str]],
        response: ResponseT,
        keys_in_command: List[KeyT],
    ):
        """
        Set a valkey command and its response in the cache.

        Args:
            command (Union[str, Sequence[str]]): The valkey command.
            response (ResponseT): The response associated with the command.
            keys_in_command (List[KeyT]): The list of keys used in the command.
        """
        if len(self.cache) >= self.max_size:
            self._evict()
        self.cache[command] = {
            _RESPONSE: response,
            _KEYS: keys_in_command,
            _CTIME: time.monotonic(),
            _ACCESS_COUNT: 0,  # Used only for LFU
        }
        self._update_key_commands_map(keys_in_command, command)
        self.commands_ttl_list.append(command)

    def get(self, command: Union[str, Sequence[str]]) -> ResponseT:
        """
        Get the response for a valkey command from the cache.

        Args:
            command (Union[str, Sequence[str]]): The valkey command.

        Returns:
            ResponseT: The response associated with the command, or None if the command is not in the cache.  # noqa
        """
        if command in self.cache:
            if self._is_expired(command):
                self.delete_command(command)
                return
            self._update_access(command)
            return copy.deepcopy(self.cache[command]["response"])

    def delete_command(self, command: Union[str, Sequence[str]]):
        """
        Delete a valkey command and its metadata from the cache.

        Args:
            command (Union[str, Sequence[str]]): The valkey command to be deleted.
        """
        if command in self.cache:
            keys_in_command = self.cache[command].get("keys")
            self._del_key_commands_map(keys_in_command, command)
            self.commands_ttl_list.remove(command)
            del self.cache[command]

    def delete_commands(self, commands: List[Union[str, Sequence[str]]]):
        """
        Delete multiple commands and their metadata from the cache.

        Args:
            commands (List[Union[str, Sequence[str]]]): The list of commands to be
            deleted.
        """
        for command in commands:
            self.delete_command(command)

    def flush(self):
        """Clear the entire cache, removing all valkey commands and metadata."""
        self.cache.clear()
        self.key_commands_map.clear()
        self.commands_ttl_list = []

    def _is_expired(self, command: Union[str, Sequence[str]]) -> bool:
        """
        Check if a valkey command has expired based on its time-to-live.

        Args:
            command (Union[str, Sequence[str]]): The valkey command.

        Returns:
            bool: True if the command has expired, False otherwise.
        """
        if self.ttl == 0:
            return False
        return time.monotonic() - self.cache[command]["ctime"] > self.ttl

    def _update_access(self, command: Union[str, Sequence[str]]):
        """
        Update the access information for a valkey command based on the eviction policy.

        Args:
            command (Union[str, Sequence[str]]): The valkey command.
        """
        if self.eviction_policy == EvictionPolicy.LRU:
            self.cache.move_to_end(command)
        elif self.eviction_policy == EvictionPolicy.LFU:
            self.cache[command]["access_count"] = (
                self.cache.get(command, {}).get("access_count", 0) + 1
            )
            self.cache.move_to_end(command)
        elif self.eviction_policy == EvictionPolicy.RANDOM:
            pass  # Random eviction doesn't require updates

    def _evict(self):
        """Evict a valkey command from the cache based on the eviction policy."""
        if self._is_expired(self.commands_ttl_list[0]):
            self.delete_command(self.commands_ttl_list[0])
        elif self.eviction_policy == EvictionPolicy.LRU:
            self.cache.popitem(last=False)
        elif self.eviction_policy == EvictionPolicy.LFU:
            min_access_command = min(
                self.cache, key=lambda k: self.cache[k].get("access_count", 0)
            )
            self.cache.pop(min_access_command)
        elif self.eviction_policy == EvictionPolicy.RANDOM:
            random_command = random.choice(list(self.cache.keys()))
            self.cache.pop(random_command)

    def _update_key_commands_map(
        self, keys: List[KeyT], command: Union[str, Sequence[str]]
    ):
        """
        Update the key_commands_map with command that uses the keys.

        Args:
            keys (List[KeyT]): The list of keys used in the command.
            command (Union[str, Sequence[str]]): The valkey command.
        """
        for key in keys:
            self.key_commands_map[key].add(command)

    def _del_key_commands_map(
        self, keys: List[KeyT], command: Union[str, Sequence[str]]
    ):
        """
        Remove a valkey command from the key_commands_map.

        Args:
            keys (List[KeyT]): The list of keys used in the valkey command.
            command (Union[str, Sequence[str]]): The valkey command.
        """
        for key in keys:
            self.key_commands_map[key].remove(command)

    def invalidate_key(self, key: KeyT):
        """
        Invalidate (delete) all valkey commands associated with a specific key.

        Args:
            key (KeyT): The key to be invalidated.
        """
        if key not in self.key_commands_map:
            return
        commands = list(self.key_commands_map[key])
        for command in commands:
            self.delete_command(command)



import time
from collections import defaultdict
from typing import List, Sequence, Union

import cachetools
import pytest
import valkey
from tests.conftest import _get_client
from valkey import ValkeyError
from valkey._cache import AbstractCache, EvictionPolicy, _LocalCache
from valkey.typing import KeyT, ResponseT
from valkey.utils import LIBVALKEY_AVAILABLE


@pytest.fixture()
def r(request):
    cache = request.param.get("cache")
    kwargs = request.param.get("kwargs", {})
    protocol = request.param.get("protocol", 3)
    single_connection_client = request.param.get("single_connection_client", False)
    with _get_client(
        valkey.Valkey,
        request,
        single_connection_client=single_connection_client,
        protocol=protocol,
        client_cache=cache,
        **kwargs,
    ) as client:
        yield client, cache


@pytest.fixture()
def local_cache():
    return _LocalCache()


@pytest.mark.skipif(LIBVALKEY_AVAILABLE, reason="PythonParser only")
class TestLocalCache:
    @pytest.mark.parametrize("r", [{"cache": _LocalCache()}], indirect=True)
    @pytest.mark.onlynoncluster
    def test_get_from_cache(self, r, r2):
        r, cache = r
        # add key to valkey
        r.set("foo", "bar")
        # get key from valkey and save in local cache
        assert r.get("foo") == b"bar"
        # get key from local cache
        assert cache.get(("GET", "foo")) == b"bar"
        # change key in valkey (cause invalidation)
        r2.set("foo", "barbar")
        # send any command to valkey (process invalidation in background)
        r.ping()
        # the command is not in the local cache anymore
        assert cache.get(("GET", "foo")) is None
        # get key from valkey
        assert r.get("foo") == b"barbar"

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(max_size=3)}],
        indirect=True,
    )
    def test_cache_lru_eviction(self, r):
        r, cache = r
        # add 3 keys to valkey
        r.set("foo", "bar")
        r.set("foo2", "bar2")
        r.set("foo3", "bar3")
        # get 3 keys from valkey and save in local cache
        assert r.get("foo") == b"bar"
        assert r.get("foo2") == b"bar2"
        assert r.get("foo3") == b"bar3"
        # get the 3 keys from local cache
        assert cache.get(("GET", "foo")) == b"bar"
        assert cache.get(("GET", "foo2")) == b"bar2"
        assert cache.get(("GET", "foo3")) == b"bar3"
        # add 1 more key to valkey (exceed the max size)
        r.set("foo4", "bar4")
        assert r.get("foo4") == b"bar4"
        # the first key is not in the local cache anymore
        assert cache.get(("GET", "foo")) is None

    @pytest.mark.parametrize("r", [{"cache": _LocalCache(ttl=1)}], indirect=True)
    def test_cache_ttl(self, r):
        r, cache = r
        # add key to valkey
        r.set("foo", "bar")
        # get key from valkey and save in local cache
        assert r.get("foo") == b"bar"
        # get key from local cache
        assert cache.get(("GET", "foo")) == b"bar"
        # wait for the key to expire
        time.sleep(1)
        # the key is not in the local cache anymore
        assert cache.get(("GET", "foo")) is None

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(max_size=3, eviction_policy=EvictionPolicy.LFU)}],
        indirect=True,
    )
    def test_cache_lfu_eviction(self, r):
        r, cache = r
        # add 3 keys to valkey
        r.set("foo", "bar")
        r.set("foo2", "bar2")
        r.set("foo3", "bar3")
        # get 3 keys from valkey and save in local cache
        assert r.get("foo") == b"bar"
        assert r.get("foo2") == b"bar2"
        assert r.get("foo3") == b"bar3"
        # change the order of the keys in the cache
        assert cache.get(("GET", "foo")) == b"bar"
        assert cache.get(("GET", "foo")) == b"bar"
        assert cache.get(("GET", "foo3")) == b"bar3"
        # add 1 more key to valkey (exceed the max size)
        r.set("foo4", "bar4")
        assert r.get("foo4") == b"bar4"
        # test the eviction policy
        assert len(cache.cache) == 3
        assert cache.get(("GET", "foo")) == b"bar"
        assert cache.get(("GET", "foo2")) is None

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    @pytest.mark.onlynoncluster
    def test_cache_decode_response(self, r):
        r, cache = r
        r.set("foo", "bar")
        # get key from valkey and save in local cache
        assert r.get("foo") == "bar"
        # get key from local cache
        assert cache.get(("GET", "foo")) == "bar"
        # change key in valkey (cause invalidation)
        r.set("foo", "barbar")
        # send any command to valkey (process invalidation in background)
        r.ping()
        # the command is not in the local cache anymore
        assert cache.get(("GET", "foo")) is None
        # get key from valkey
        assert r.get("foo") == "barbar"

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"cache_deny_list": ["LLEN"]}}],
        indirect=True,
    )
    def test_cache_deny_list(self, r):
        r, cache = r
        # add list to valkey
        r.lpush("mylist", "foo", "bar", "baz")
        assert r.llen("mylist") == 3
        assert r.lindex("mylist", 1) == b"bar"
        assert cache.get(("LLEN", "mylist")) is None
        assert cache.get(("LINDEX", "mylist", 1)) == b"bar"

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"cache_allow_list": ["LLEN"]}}],
        indirect=True,
    )
    def test_cache_allow_list(self, r):
        r, cache = r
        r.lpush("mylist", "foo", "bar", "baz")
        assert r.llen("mylist") == 3
        assert r.lindex("mylist", 1) == b"bar"
        assert cache.get(("LLEN", "mylist")) == 3
        assert cache.get(("LINDEX", "mylist", 1)) is None

    @pytest.mark.parametrize("r", [{"cache": _LocalCache()}], indirect=True)
    def test_cache_return_copy(self, r):
        r, cache = r
        r.lpush("mylist", "foo", "bar", "baz")
        assert r.lrange("mylist", 0, -1) == [b"baz", b"bar", b"foo"]
        res = cache.get(("LRANGE", "mylist", 0, -1))
        assert res == [b"baz", b"bar", b"foo"]
        res.append(b"new")
        check = cache.get(("LRANGE", "mylist", 0, -1))
        assert check == [b"baz", b"bar", b"foo"]

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    @pytest.mark.onlynoncluster
    def test_csc_not_cause_disconnects(self, r):
        r, cache = r
        id1 = r.client_id()
        r.mset({"a": 1, "b": 1, "c": 1, "d": 1, "e": 1, "f": 1})
        assert r.mget("a", "b", "c", "d", "e", "f") == ["1", "1", "1", "1", "1", "1"]
        id2 = r.client_id()

        # client should get value from client cache
        assert r.mget("a", "b", "c", "d", "e", "f") == ["1", "1", "1", "1", "1", "1"]
        assert cache.get(("MGET", "a", "b", "c", "d", "e", "f")) == [
            "1",
            "1",
            "1",
            "1",
            "1",
            "1",
        ]

        r.mset({"a": 2, "b": 2, "c": 2, "d": 2, "e": 2, "f": 2})
        id3 = r.client_id()
        # client should get value from valkey server post invalidate messages
        assert r.mget("a", "b", "c", "d", "e", "f") == ["2", "2", "2", "2", "2", "2"]

        r.mset({"a": 3, "b": 3, "c": 3, "d": 3, "e": 3, "f": 3})
        # need to check that we get correct value 3 and not 2
        assert r.mget("a", "b", "c", "d", "e", "f") == ["3", "3", "3", "3", "3", "3"]
        # client should get value from client cache
        assert r.mget("a", "b", "c", "d", "e", "f") == ["3", "3", "3", "3", "3", "3"]

        r.mset({"a": 4, "b": 4, "c": 4, "d": 4, "e": 4, "f": 4})
        # need to check that we get correct value 4 and not 3
        assert r.mget("a", "b", "c", "d", "e", "f") == ["4", "4", "4", "4", "4", "4"]
        # client should get value from client cache
        assert r.mget("a", "b", "c", "d", "e", "f") == ["4", "4", "4", "4", "4", "4"]
        id4 = r.client_id()
        assert id1 == id2 == id3 == id4

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    @pytest.mark.onlynoncluster
    def test_multiple_commands_same_key(self, r):
        r, cache = r
        r.mset({"a": 1, "b": 1})
        assert r.mget("a", "b") == ["1", "1"]
        # value should be in local cache
        assert cache.get(("MGET", "a", "b")) == ["1", "1"]
        # set only one key
        r.set("a", 2)
        # send any command to valkey (process invalidation in background)
        r.ping()
        # the command is not in the local cache anymore
        assert cache.get(("MGET", "a", "b")) is None
        # get from valkey
        assert r.mget("a", "b") == ["2", "1"]

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    def test_delete_one_command(self, r):
        r, cache = r
        r.mset({"a{a}": 1, "b{a}": 1})
        r.set("c", 1)
        assert r.mget("a{a}", "b{a}") == ["1", "1"]
        assert r.get("c") == "1"
        # values should be in local cache
        assert cache.get(("MGET", "a{a}", "b{a}")) == ["1", "1"]
        assert cache.get(("GET", "c")) == "1"
        # delete one command from the cache
        r.delete_command_from_cache(("MGET", "a{a}", "b{a}"))
        # the other command is still in the local cache anymore
        assert cache.get(("MGET", "a{a}", "b{a}")) is None
        assert cache.get(("GET", "c")) == "1"
        # get from valkey
        assert r.mget("a{a}", "b{a}") == ["1", "1"]
        assert r.get("c") == "1"

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    def test_delete_several_commands(self, r):
        r, cache = r
        r.mset({"a{a}": 1, "b{a}": 1})
        r.set("c", 1)
        assert r.mget("a{a}", "b{a}") == ["1", "1"]
        assert r.get("c") == "1"
        # values should be in local cache
        assert cache.get(("MGET", "a{a}", "b{a}")) == ["1", "1"]
        assert cache.get(("GET", "c")) == "1"
        # delete the commands from the cache
        cache.delete_commands([("MGET", "a{a}", "b{a}"), ("GET", "c")])
        # the commands are not in the local cache anymore
        assert cache.get(("MGET", "a{a}", "b{a}")) is None
        assert cache.get(("GET", "c")) is None
        # get from valkey
        assert r.mget("a{a}", "b{a}") == ["1", "1"]
        assert r.get("c") == "1"

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    def test_invalidate_key(self, r):
        r, cache = r
        r.mset({"a{a}": 1, "b{a}": 1})
        r.set("c", 1)
        assert r.mget("a{a}", "b{a}") == ["1", "1"]
        assert r.get("c") == "1"
        # values should be in local cache
        assert cache.get(("MGET", "a{a}", "b{a}")) == ["1", "1"]
        assert cache.get(("GET", "c")) == "1"
        # invalidate one key from the cache
        r.invalidate_key_from_cache("b{a}")
        # one other command is still in the local cache anymore
        assert cache.get(("MGET", "a{a}", "b{a}")) is None
        assert cache.get(("GET", "c")) == "1"
        # get from valkey
        assert r.mget("a{a}", "b{a}") == ["1", "1"]
        assert r.get("c") == "1"

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    def test_flush_entire_cache(self, r):
        r, cache = r
        r.mset({"a{a}": 1, "b{a}": 1})
        r.set("c", 1)
        assert r.mget("a{a}", "b{a}") == ["1", "1"]
        assert r.get("c") == "1"
        # values should be in local cache
        assert cache.get(("MGET", "a{a}", "b{a}")) == ["1", "1"]
        assert cache.get(("GET", "c")) == "1"
        # flush the local cache
        r.flush_cache()
        # the commands are not in the local cache anymore
        assert cache.get(("MGET", "a{a}", "b{a}")) is None
        assert cache.get(("GET", "c")) is None
        # get from valkey
        assert r.mget("a{a}", "b{a}") == ["1", "1"]
        assert r.get("c") == "1"

    @pytest.mark.onlynoncluster
    def test_cache_not_available_with_resp2(self, request):
        with pytest.raises(ValkeyError) as e:
            _get_client(valkey.Valkey, request, protocol=2, client_cache=_LocalCache())
        assert "protocol version 3 or higher" in str(e.value)

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    @pytest.mark.onlynoncluster
    def test_execute_command_args_not_split(self, r):
        r, cache = r
        assert r.execute_command("SET a 1") == "OK"
        assert r.execute_command("GET a") == "1"
        # "get a" is not whitelisted by default, the args should be separated
        assert cache.get(("GET a",)) is None

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    def test_execute_command_keys_provided(self, r):
        r, cache = r
        assert r.execute_command("SET", "b", "2") is True
        assert r.execute_command("GET", "b", keys=["b"]) == "2"
        assert cache.get(("GET", "b")) == "2"

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    def test_execute_command_keys_not_provided(self, r):
        r, cache = r
        assert r.execute_command("SET", "b", "2") is True
        assert r.execute_command("GET", "b") == "2"  # keys not provided, not cached
        assert cache.get(("GET", "b")) is None

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "single_connection_client": True}],
        indirect=True,
    )
    @pytest.mark.onlynoncluster
    def test_single_connection(self, r):
        r, cache = r
        # add key to valkey
        r.set("foo", "bar")
        # get key from valkey and save in local cache
        assert r.get("foo") == b"bar"
        # get key from local cache
        assert cache.get(("GET", "foo")) == b"bar"
        # change key in valkey (cause invalidation)
        r.set("foo", "barbar")
        # send any command to valkey (process invalidation in background)
        r.ping()
        # the command is not in the local cache anymore
        assert cache.get(("GET", "foo")) is None
        # get key from valkey
        assert r.get("foo") == b"barbar"

    @pytest.mark.parametrize("r", [{"cache": _LocalCache()}], indirect=True)
    def test_get_from_cache_invalidate_via_get(self, r, r2):
        r, cache = r
        # add key to valkey
        r.set("foo", "bar")
        # get key from valkey and save in local cache
        assert r.get("foo") == b"bar"
        # get key from local cache
        assert cache.get(("GET", "foo")) == b"bar"
        # change key in valkey (cause invalidation)
        r2.set("foo", "barbar")
        # don't send any command to valkey, just run another get
        # it should process the invalidation in background
        assert r.get("foo") == b"barbar"


@pytest.mark.skipif(LIBVALKEY_AVAILABLE, reason="PythonParser only")
@pytest.mark.onlycluster
class TestClusterLocalCache:
    @pytest.mark.parametrize("r", [{"cache": _LocalCache()}], indirect=True)
    def test_get_from_cache(self, r, r2):
        r, cache = r
        # add key to valkey
        r.set("foo", "bar")
        # get key from valkey and save in local cache
        assert r.get("foo") == b"bar"
        # get key from local cache
        assert cache.get(("GET", "foo")) == b"bar"
        # change key in valkey (cause invalidation)
        r2.set("foo", "barbar")
        # send any command to valkey (process invalidation in background)
        node = r.get_node_from_key("foo")
        r.ping(target_nodes=node)
        # the command is not in the local cache anymore
        assert cache.get(("GET", "foo")) is None
        # get key from valkey
        assert r.get("foo") == b"barbar"

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    def test_cache_decode_response(self, r):
        r, cache = r
        r.set("foo", "bar")
        # get key from valkey and save in local cache
        assert r.get("foo") == "bar"
        # get key from local cache
        assert cache.get(("GET", "foo")) == "bar"
        # change key in valkey (cause invalidation)
        r.set("foo", "barbar")
        # send any command to valkey (process invalidation in background)
        node = r.get_node_from_key("foo")
        r.ping(target_nodes=node)
        # the command is not in the local cache anymore
        assert cache.get(("GET", "foo")) is None
        # get key from valkey
        assert r.get("foo") == "barbar"

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    def test_execute_command_keys_provided(self, r):
        r, cache = r
        assert r.execute_command("SET", "b", "2") is True
        assert r.execute_command("GET", "b", keys=["b"]) == "2"
        assert cache.get(("GET", "b")) == "2"

    @pytest.mark.parametrize(
        "r",
        [{"cache": _LocalCache(), "kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    def test_execute_command_keys_not_provided(self, r):
        r, cache = r
        assert r.execute_command("SET", "b", "2") is True
        assert r.execute_command("GET", "b") == "2"  # keys not provided, not cached
        assert cache.get(("GET", "b")) is None


@pytest.mark.skipif(LIBVALKEY_AVAILABLE, reason="PythonParser only")
@pytest.mark.onlynoncluster
class TestSentinelLocalCache:

    def test_get_from_cache(self, local_cache, master):
        master.set("foo", "bar")
        # get key from valkey and save in local cache
        assert master.get("foo") == b"bar"
        # get key from local cache
        assert local_cache.get(("GET", "foo")) == b"bar"
        # change key in valkey (cause invalidation)
        master.set("foo", "barbar")
        # send any command to valkey (process invalidation in background)
        master.ping()
        # the command is not in the local cache anymore
        assert local_cache.get(("GET", "foo")) is None
        # get key from valkey
        assert master.get("foo") == b"barbar"

    @pytest.mark.parametrize(
        "sentinel_setup",
        [{"kwargs": {"decode_responses": True}}],
        indirect=True,
    )
    def test_cache_decode_response(self, local_cache, sentinel_setup, master):
        master.set("foo", "bar")
        # get key from valkey and save in local cache
        assert master.get("foo") == "bar"
        # get key from local cache
        assert local_cache.get(("GET", "foo")) == "bar"
        # change key in valkey (cause invalidation)
        master.set("foo", "barbar")
        # send any command to valkey (process invalidation in background)
        master.ping()
        # the command is not in the local cache anymore
        assert local_cache.get(("GET", "foo")) is None
        # get key from valkey
        assert master.get("foo") == "barbar"


@pytest.mark.skipif(LIBVALKEY_AVAILABLE, reason="PythonParser only")
@pytest.mark.onlynoncluster
class TestCustomCache:
    class _CustomCache(AbstractCache):
        def __init__(self):
            self.responses = cachetools.LRUCache(maxsize=1000)
            self.keys_to_commands = defaultdict(list)
            self.commands_to_keys = defaultdict(list)

        def set(
            self,
            command: Union[str, Sequence[str]],
            response: ResponseT,
            keys_in_command: List[KeyT],
        ):
            self.responses[command] = response
            for key in keys_in_command:
                self.keys_to_commands[key].append(tuple(command))
                self.commands_to_keys[command].append(tuple(keys_in_command))

        def get(self, command: Union[str, Sequence[str]]) -> ResponseT:
            return self.responses.get(command)

        def delete_command(self, command: Union[str, Sequence[str]]):
            self.responses.pop(command, None)
            keys = self.commands_to_keys.pop(command, [])
            for key in keys:
                if command in self.keys_to_commands[key]:
                    self.keys_to_commands[key].remove(command)

        def delete_commands(self, commands: List[Union[str, Sequence[str]]]):
            for command in commands:
                self.delete_command(command)

        def flush(self):
            self.responses.clear()
            self.commands_to_keys.clear()
            self.keys_to_commands.clear()

        def invalidate_key(self, key: KeyT):
            commands = self.keys_to_commands.pop(key, [])
            for command in commands:
                self.delete_command(command)

    @pytest.mark.parametrize("r", [{"cache": _CustomCache()}], indirect=True)
    def test_get_from_cache(self, r, r2):
        r, cache = r
        # add key to valkey
        r.set("foo", "bar")
        # get key from valkey and save in local cache
        assert r.get("foo") == b"bar"
        # get key from local cache
        assert cache.get(("GET", "foo")) == b"bar"
        # change key in valkey (cause invalidation)
        r2.set("foo", "barbar")
        # send any command to valkey (process invalidation in background)
        r.ping()
        # the command is not in the local cache anymore
        assert cache.get(("GET", "foo")) is None
        # get key from valkey
        assert r.get("foo") == b"barbar"



import os
import re
import time
from contextlib import closing
from threading import Thread
from unittest import mock

import pytest
import valkey
from valkey._parsers.url_parser import to_bool
from valkey.utils import SSL_AVAILABLE

from .conftest import _get_client, skip_if_server_version_lt
from .test_pubsub import wait_for_message


class DummyConnection:
    description_format = "DummyConnection<>"

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.pid = os.getpid()

    def connect(self):
        pass

    def can_read(self):
        return False


class TestConnectionPool:
    def get_pool(
        self,
        connection_kwargs=None,
        max_connections=None,
        connection_class=valkey.Connection,
    ):
        connection_kwargs = connection_kwargs or {}
        pool = valkey.ConnectionPool(
            connection_class=connection_class,
            max_connections=max_connections,
            **connection_kwargs,
        )
        return pool

    def test_connection_creation(self):
        connection_kwargs = {"foo": "bar", "biz": "baz"}
        pool = self.get_pool(
            connection_kwargs=connection_kwargs, connection_class=DummyConnection
        )
        connection = pool.get_connection("_")
        assert isinstance(connection, DummyConnection)
        assert connection.kwargs == connection_kwargs

    def test_closing(self):
        connection_kwargs = {"foo": "bar", "biz": "baz"}
        pool = valkey.ConnectionPool(
            connection_class=DummyConnection,
            max_connections=None,
            **connection_kwargs,
        )
        with closing(pool):
            pass

    def test_multiple_connections(self, master_host):
        connection_kwargs = {"host": master_host[0], "port": master_host[1]}
        pool = self.get_pool(connection_kwargs=connection_kwargs)
        c1 = pool.get_connection("_")
        c2 = pool.get_connection("_")
        assert c1 != c2

    def test_max_connections(self, master_host):
        connection_kwargs = {"host": master_host[0], "port": master_host[1]}
        pool = self.get_pool(max_connections=2, connection_kwargs=connection_kwargs)
        pool.get_connection("_")
        pool.get_connection("_")
        with pytest.raises(valkey.ConnectionError):
            pool.get_connection("_")

    def test_reuse_previously_released_connection(self, master_host):
        connection_kwargs = {"host": master_host[0], "port": master_host[1]}
        pool = self.get_pool(connection_kwargs=connection_kwargs)
        c1 = pool.get_connection("_")
        pool.release(c1)
        c2 = pool.get_connection("_")
        assert c1 == c2

    def test_repr_contains_db_info_tcp(self):
        connection_kwargs = {
            "host": "localhost",
            "port": 6379,
            "db": 1,
            "client_name": "test-client",
        }
        pool = self.get_pool(
            connection_kwargs=connection_kwargs, connection_class=valkey.Connection
        )
        expected = "host=localhost,port=6379,db=1,client_name=test-client"
        assert expected in repr(pool)

    def test_repr_contains_db_info_unix(self):
        connection_kwargs = {"path": "/abc", "db": 1, "client_name": "test-client"}
        pool = self.get_pool(
            connection_kwargs=connection_kwargs,
            connection_class=valkey.UnixDomainSocketConnection,
        )
        expected = "path=/abc,db=1,client_name=test-client"
        assert expected in repr(pool)


class TestBlockingConnectionPool:
    def get_pool(self, connection_kwargs=None, max_connections=10, timeout=20):
        connection_kwargs = connection_kwargs or {}
        pool = valkey.BlockingConnectionPool(
            connection_class=DummyConnection,
            max_connections=max_connections,
            timeout=timeout,
            **connection_kwargs,
        )
        return pool

    def test_connection_creation(self, master_host):
        connection_kwargs = {
            "foo": "bar",
            "biz": "baz",
            "host": master_host[0],
            "port": master_host[1],
        }
        pool = self.get_pool(connection_kwargs=connection_kwargs)
        connection = pool.get_connection("_")
        assert isinstance(connection, DummyConnection)
        assert connection.kwargs == connection_kwargs

    def test_multiple_connections(self, master_host):
        connection_kwargs = {"host": master_host[0], "port": master_host[1]}
        pool = self.get_pool(connection_kwargs=connection_kwargs)
        c1 = pool.get_connection("_")
        c2 = pool.get_connection("_")
        assert c1 != c2

    def test_connection_pool_blocks_until_timeout(self, master_host):
        "When out of connections, block for timeout seconds, then raise"
        connection_kwargs = {"host": master_host[0], "port": master_host[1]}
        pool = self.get_pool(
            max_connections=1, timeout=0.1, connection_kwargs=connection_kwargs
        )
        pool.get_connection("_")

        start = time.time()
        with pytest.raises(valkey.ConnectionError):
            pool.get_connection("_")
        # we should have waited at least 0.1 seconds
        assert time.time() - start >= 0.1

    def test_connection_pool_blocks_until_conn_available(self, master_host):
        """
        When out of connections, block until another connection is released
        to the pool
        """
        connection_kwargs = {"host": master_host[0], "port": master_host[1]}
        pool = self.get_pool(
            max_connections=1, timeout=2, connection_kwargs=connection_kwargs
        )
        c1 = pool.get_connection("_")

        def target():
            time.sleep(0.1)
            pool.release(c1)

        start = time.time()
        Thread(target=target).start()
        pool.get_connection("_")
        assert time.time() - start >= 0.1

    def test_reuse_previously_released_connection(self, master_host):
        connection_kwargs = {"host": master_host[0], "port": master_host[1]}
        pool = self.get_pool(connection_kwargs=connection_kwargs)
        c1 = pool.get_connection("_")
        pool.release(c1)
        c2 = pool.get_connection("_")
        assert c1 == c2

    def test_repr_contains_db_info_tcp(self):
        pool = valkey.ConnectionPool(
            host="localhost", port=6379, client_name="test-client"
        )
        expected = "host=localhost,port=6379,db=0,client_name=test-client"
        assert expected in repr(pool)

    def test_repr_contains_db_info_unix(self):
        pool = valkey.ConnectionPool(
            connection_class=valkey.UnixDomainSocketConnection,
            path="abc",
            client_name="test-client",
        )
        expected = "path=abc,db=0,client_name=test-client"
        assert expected in repr(pool)


class TestConnectionPoolURLParsing:
    def test_hostname(self):
        pool = valkey.ConnectionPool.from_url("valkey://my.host")
        assert pool.connection_class == valkey.Connection
        assert pool.connection_kwargs == {"host": "my.host"}

    def test_quoted_hostname(self):
        pool = valkey.ConnectionPool.from_url("valkey://my %2F host %2B%3D+")
        assert pool.connection_class == valkey.Connection
        assert pool.connection_kwargs == {"host": "my / host +=+"}

    def test_port(self):
        pool = valkey.ConnectionPool.from_url("valkey://localhost:6380")
        assert pool.connection_class == valkey.Connection
        assert pool.connection_kwargs == {"host": "localhost", "port": 6380}

    @skip_if_server_version_lt("6.0.0")
    def test_username(self):
        pool = valkey.ConnectionPool.from_url("valkey://myuser:@localhost")
        assert pool.connection_class == valkey.Connection
        assert pool.connection_kwargs == {"host": "localhost", "username": "myuser"}

    @skip_if_server_version_lt("6.0.0")
    def test_quoted_username(self):
        pool = valkey.ConnectionPool.from_url(
            "valkey://%2Fmyuser%2F%2B name%3D%24+:@localhost"
        )
        assert pool.connection_class == valkey.Connection
        assert pool.connection_kwargs == {
            "host": "localhost",
            "username": "/myuser/+ name=$+",
        }

    def test_password(self):
        pool = valkey.ConnectionPool.from_url("valkey://:mypassword@localhost")
        assert pool.connection_class == valkey.Connection
        assert pool.connection_kwargs == {"host": "localhost", "password": "mypassword"}

    def test_quoted_password(self):
        pool = valkey.ConnectionPool.from_url(
            "valkey://:%2Fmypass%2F%2B word%3D%24+@localhost"
        )
        assert pool.connection_class == valkey.Connection
        assert pool.connection_kwargs == {
            "host": "localhost",
            "password": "/mypass/+ word=$+",
        }

    @skip_if_server_version_lt("6.0.0")
    def test_username_and_password(self):
        pool = valkey.ConnectionPool.from_url("valkey://myuser:mypass@localhost")
        assert pool.connection_class == valkey.Connection
        assert pool.connection_kwargs == {
            "host": "localhost",
            "username": "myuser",
            "password": "mypass",
        }

    def test_db_as_argument(self):
        pool = valkey.ConnectionPool.from_url("valkey://localhost", db=1)
        assert pool.connection_class == valkey.Connection
        assert pool.connection_kwargs == {"host": "localhost", "db": 1}

    def test_db_in_path(self):
        pool = valkey.ConnectionPool.from_url("valkey://localhost/2", db=1)
        assert pool.connection_class == valkey.Connection
        assert pool.connection_kwargs == {"host": "localhost", "db": 2}

    def test_db_in_querystring(self):
        pool = valkey.ConnectionPool.from_url("valkey://localhost/2?db=3", db=1)
        assert pool.connection_class == valkey.Connection
        assert pool.connection_kwargs == {"host": "localhost", "db": 3}

    def test_extra_typed_querystring_options(self):
        pool = valkey.ConnectionPool.from_url(
            "valkey://localhost/2?socket_timeout=20&socket_connect_timeout=10"
            "&socket_keepalive=&retry_on_timeout=Yes&max_connections=10"
        )

        assert pool.connection_class == valkey.Connection
        assert pool.connection_kwargs == {
            "host": "localhost",
            "db": 2,
            "socket_timeout": 20.0,
            "socket_connect_timeout": 10.0,
            "retry_on_timeout": True,
        }
        assert pool.max_connections == 10

    def test_boolean_parsing(self):
        for expected, value in (
            (None, None),
            (None, ""),
            (False, 0),
            (False, "0"),
            (False, "f"),
            (False, "F"),
            (False, "False"),
            (False, "n"),
            (False, "N"),
            (False, "No"),
            (True, 1),
            (True, "1"),
            (True, "y"),
            (True, "Y"),
            (True, "Yes"),
        ):
            assert expected is to_bool(value)

    def test_client_name_in_querystring(self):
        pool = valkey.ConnectionPool.from_url(
            "valkey://location?client_name=test-client"
        )
        assert pool.connection_kwargs["client_name"] == "test-client"

    def test_invalid_extra_typed_querystring_options(self):
        with pytest.raises(ValueError):
            valkey.ConnectionPool.from_url(
                "valkey://localhost/2?socket_timeout=_&socket_connect_timeout=abc"
            )

    def test_extra_querystring_options(self):
        pool = valkey.ConnectionPool.from_url("valkey://localhost?a=1&b=2")
        assert pool.connection_class == valkey.Connection
        assert pool.connection_kwargs == {"host": "localhost", "a": "1", "b": "2"}

    def test_calling_from_subclass_returns_correct_instance(self):
        pool = valkey.BlockingConnectionPool.from_url("valkey://localhost")
        assert isinstance(pool, valkey.BlockingConnectionPool)

    def test_client_creates_connection_pool(self):
        r = valkey.Valkey.from_url("valkey://myhost")
        assert r.connection_pool.connection_class == valkey.Connection
        assert r.connection_pool.connection_kwargs == {"host": "myhost"}

    def test_invalid_scheme_raises_error(self):
        with pytest.raises(ValueError) as cm:
            valkey.ConnectionPool.from_url("localhost")
        assert str(cm.value) == (
            "Valkey URL must specify one of the following schemes "
            "['valkey', 'valkeys', 'redis', 'rediss', 'unix']"
        )

    def test_invalid_scheme_raises_error_when_double_slash_missing(self):
        with pytest.raises(ValueError) as cm:
            valkey.ConnectionPool.from_url("valkey:foo.bar.com:12345")
        assert str(cm.value) == (
            "Valkey URL must specify one of the following schemes "
            "['valkey', 'valkeys', 'redis', 'rediss', 'unix']"
        )


class TestBlockingConnectionPoolURLParsing:
    def test_extra_typed_querystring_options(self):
        pool = valkey.BlockingConnectionPool.from_url(
            "valkey://localhost/2?socket_timeout=20&socket_connect_timeout=10"
            "&socket_keepalive=&retry_on_timeout=Yes&max_connections=10&timeout=42"
        )

        assert pool.connection_class == valkey.Connection
        assert pool.connection_kwargs == {
            "host": "localhost",
            "db": 2,
            "socket_timeout": 20.0,
            "socket_connect_timeout": 10.0,
            "retry_on_timeout": True,
        }
        assert pool.max_connections == 10
        assert pool.timeout == 42.0

    def test_invalid_extra_typed_querystring_options(self):
        with pytest.raises(ValueError):
            valkey.BlockingConnectionPool.from_url(
                "valkey://localhost/2?timeout=_not_a_float_"
            )


class TestConnectionPoolUnixSocketURLParsing:
    def test_defaults(self):
        pool = valkey.ConnectionPool.from_url("unix:///socket")
        assert pool.connection_class == valkey.UnixDomainSocketConnection
        assert pool.connection_kwargs == {"path": "/socket"}

    @skip_if_server_version_lt("6.0.0")
    def test_username(self):
        pool = valkey.ConnectionPool.from_url("unix://myuser:@/socket")
        assert pool.connection_class == valkey.UnixDomainSocketConnection
        assert pool.connection_kwargs == {"path": "/socket", "username": "myuser"}

    @skip_if_server_version_lt("6.0.0")
    def test_quoted_username(self):
        pool = valkey.ConnectionPool.from_url(
            "unix://%2Fmyuser%2F%2B name%3D%24+:@/socket"
        )
        assert pool.connection_class == valkey.UnixDomainSocketConnection
        assert pool.connection_kwargs == {
            "path": "/socket",
            "username": "/myuser/+ name=$+",
        }

    def test_password(self):
        pool = valkey.ConnectionPool.from_url("unix://:mypassword@/socket")
        assert pool.connection_class == valkey.UnixDomainSocketConnection
        assert pool.connection_kwargs == {"path": "/socket", "password": "mypassword"}

    def test_quoted_password(self):
        pool = valkey.ConnectionPool.from_url(
            "unix://:%2Fmypass%2F%2B word%3D%24+@/socket"
        )
        assert pool.connection_class == valkey.UnixDomainSocketConnection
        assert pool.connection_kwargs == {
            "path": "/socket",
            "password": "/mypass/+ word=$+",
        }

    def test_quoted_path(self):
        pool = valkey.ConnectionPool.from_url(
            "unix://:mypassword@/my%2Fpath%2Fto%2F..%2F+_%2B%3D%24ocket"
        )
        assert pool.connection_class == valkey.UnixDomainSocketConnection
        assert pool.connection_kwargs == {
            "path": "/my/path/to/../+_+=$ocket",
            "password": "mypassword",
        }

    def test_db_as_argument(self):
        pool = valkey.ConnectionPool.from_url("unix:///socket", db=1)
        assert pool.connection_class == valkey.UnixDomainSocketConnection
        assert pool.connection_kwargs == {"path": "/socket", "db": 1}

    def test_db_in_querystring(self):
        pool = valkey.ConnectionPool.from_url("unix:///socket?db=2", db=1)
        assert pool.connection_class == valkey.UnixDomainSocketConnection
        assert pool.connection_kwargs == {"path": "/socket", "db": 2}

    def test_client_name_in_querystring(self):
        pool = valkey.ConnectionPool.from_url(
            "valkey://location?client_name=test-client"
        )
        assert pool.connection_kwargs["client_name"] == "test-client"

    def test_extra_querystring_options(self):
        pool = valkey.ConnectionPool.from_url("unix:///socket?a=1&b=2")
        assert pool.connection_class == valkey.UnixDomainSocketConnection
        assert pool.connection_kwargs == {"path": "/socket", "a": "1", "b": "2"}

    def test_connection_class_override(self):
        class MyConnection(valkey.UnixDomainSocketConnection):
            pass

        pool = valkey.ConnectionPool.from_url(
            "unix:///socket", connection_class=MyConnection
        )
        assert pool.connection_class == MyConnection


@pytest.mark.skipif(not SSL_AVAILABLE, reason="SSL not installed")
class TestSSLConnectionURLParsing:
    def test_host(self):
        pool = valkey.ConnectionPool.from_url("valkeys://my.host")
        assert pool.connection_class == valkey.SSLConnection
        assert pool.connection_kwargs == {"host": "my.host"}

    def test_connection_class_override(self):
        class MyConnection(valkey.SSLConnection):
            pass

        pool = valkey.ConnectionPool.from_url(
            "valkeys://my.host", connection_class=MyConnection
        )
        assert pool.connection_class == MyConnection

    def test_cert_reqs_options(self):
        import ssl

        class DummyConnectionPool(valkey.ConnectionPool):
            def get_connection(self, *args, **kwargs):
                return self.make_connection()

        pool = DummyConnectionPool.from_url("valkeys://?ssl_cert_reqs=none")
        assert pool.get_connection("_").cert_reqs == ssl.CERT_NONE

        pool = DummyConnectionPool.from_url("valkeys://?ssl_cert_reqs=optional")
        assert pool.get_connection("_").cert_reqs == ssl.CERT_OPTIONAL

        pool = DummyConnectionPool.from_url("valkeys://?ssl_cert_reqs=required")
        assert pool.get_connection("_").cert_reqs == ssl.CERT_REQUIRED

        pool = DummyConnectionPool.from_url("valkeys://?ssl_check_hostname=False")
        assert pool.get_connection("_").check_hostname is False

        pool = DummyConnectionPool.from_url("valkeys://?ssl_check_hostname=True")
        assert pool.get_connection("_").check_hostname is True


class TestConnection:
    def test_on_connect_error(self):
        """
        An error in Connection.on_connect should disconnect from the server
        see for details: https://github.com/andymccurdy/redis-py/issues/368
        """
        # this assumes the Valkey server being tested against doesn't have
        # 9999 databases ;)
        bad_connection = valkey.Valkey(db=9999)
        # an error should be raised on connect
        with pytest.raises(valkey.ValkeyError):
            bad_connection.info()
        pool = bad_connection.connection_pool
        assert len(pool._available_connections) == 1
        assert not pool._available_connections[0]._sock

    @pytest.mark.onlynoncluster
    @skip_if_server_version_lt("2.8.8")
    def test_busy_loading_disconnects_socket(self, r):
        """
        If Valkey raises a LOADING error, the connection should be
        disconnected and a BusyLoadingError raised
        """
        with pytest.raises(valkey.BusyLoadingError):
            r.execute_command("DEBUG", "ERROR", "LOADING fake message")
        assert not r.connection._sock

    @pytest.mark.onlynoncluster
    @skip_if_server_version_lt("2.8.8")
    def test_busy_loading_from_pipeline_immediate_command(self, r):
        """
        BusyLoadingErrors should raise from Pipelines that execute a
        command immediately, like WATCH does.
        """
        pipe = r.pipeline()
        with pytest.raises(valkey.BusyLoadingError):
            pipe.immediate_execute_command("DEBUG", "ERROR", "LOADING fake message")
        pool = r.connection_pool
        assert not pipe.connection
        assert len(pool._available_connections) == 1
        assert not pool._available_connections[0]._sock

    @pytest.mark.onlynoncluster
    @skip_if_server_version_lt("2.8.8")
    def test_busy_loading_from_pipeline(self, r):
        """
        BusyLoadingErrors should be raised from a pipeline execution
        regardless of the raise_on_error flag.
        """
        pipe = r.pipeline()
        pipe.execute_command("DEBUG", "ERROR", "LOADING fake message")
        with pytest.raises(valkey.BusyLoadingError):
            pipe.execute()
        pool = r.connection_pool
        assert not pipe.connection
        assert len(pool._available_connections) == 1
        assert not pool._available_connections[0]._sock

    @skip_if_server_version_lt("2.8.8")
    def test_read_only_error(self, r):
        "READONLY errors get turned into ReadOnlyError exceptions"
        with pytest.raises(valkey.ReadOnlyError):
            r.execute_command("DEBUG", "ERROR", "READONLY blah blah")

    def test_oom_error(self, r):
        "OOM errors get turned into OutOfMemoryError exceptions"
        with pytest.raises(valkey.OutOfMemoryError):
            # note: don't use the DEBUG OOM command since it's not the same
            # as the db being full
            r.execute_command("DEBUG", "ERROR", "OOM blah blah")

    def test_connect_from_url_tcp(self):
        connection = valkey.Valkey.from_url("valkey://localhost")
        pool = connection.connection_pool

        assert re.match(
            r"< .*?([^\.]+) \( < .*?([^\.]+) \( (.+) \) > \) >", repr(pool), re.VERBOSE
        ).groups() == (
            "ConnectionPool",
            "Connection",
            "host=localhost,port=6379,db=0",
        )

    def test_connect_from_url_unix(self):
        connection = valkey.Valkey.from_url("unix:///path/to/socket")
        pool = connection.connection_pool

        assert re.match(
            r"< .*?([^\.]+) \( < .*?([^\.]+) \( (.+) \) > \) >", repr(pool), re.VERBOSE
        ).groups() == (
            "ConnectionPool",
            "UnixDomainSocketConnection",
            "path=/path/to/socket,db=0",
        )

    def test_connect_no_auth_configured(self, r):
        """
        AuthenticationError should be raised when the server is not configured with auth
        but credentials are supplied by the user.
        """
        # Server < 6
        with pytest.raises(valkey.AuthenticationError):
            r.execute_command(
                "DEBUG", "ERROR", "ERR Client sent AUTH, but no password is set"
            )

        # Server >= 6
        with pytest.raises(valkey.AuthenticationError):
            r.execute_command(
                "DEBUG",
                "ERROR",
                "ERR AUTH <password> called without any password "
                "configured for the default user. Are you sure "
                "your configuration is correct?",
            )

    def test_connect_invalid_auth_credentials_supplied(self, r):
        """
        AuthenticationError should be raised when sending invalid username/password
        """
        with pytest.raises(valkey.AuthenticationError):
            r.execute_command("DEBUG", "ERROR", "WRONGPASS")


@pytest.mark.onlynoncluster
class TestMultiConnectionClient:
    @pytest.fixture()
    def r(self, request):
        return _get_client(valkey.Valkey, request, single_connection_client=False)

    def test_multi_connection_command(self, r):
        assert not r.connection
        assert r.set("a", "123")
        assert r.get("a") == b"123"


@pytest.mark.onlynoncluster
class TestHealthCheck:
    interval = 60

    @pytest.fixture()
    def r(self, request):
        return _get_client(valkey.Valkey, request, health_check_interval=self.interval)

    def assert_interval_advanced(self, connection):
        diff = connection.next_health_check - time.time()
        assert self.interval > diff > (self.interval - 1)

    def test_health_check_runs(self, r):
        r.connection.next_health_check = time.time() - 1
        r.connection.check_health()
        self.assert_interval_advanced(r.connection)

    def test_arbitrary_command_invokes_health_check(self, r):
        # invoke a command to make sure the connection is entirely setup
        r.get("foo")
        r.connection.next_health_check = time.time()
        with mock.patch.object(
            r.connection, "send_command", wraps=r.connection.send_command
        ) as m:
            r.get("foo")
            m.assert_called_with("PING", check_health=False)

        self.assert_interval_advanced(r.connection)

    def test_arbitrary_command_advances_next_health_check(self, r):
        r.get("foo")
        next_health_check = r.connection.next_health_check
        r.get("foo")
        assert next_health_check < r.connection.next_health_check

    def test_health_check_not_invoked_within_interval(self, r):
        r.get("foo")
        with mock.patch.object(
            r.connection, "send_command", wraps=r.connection.send_command
        ) as m:
            r.get("foo")
            ping_call_spec = (("PING",), {"check_health": False})
            assert ping_call_spec not in m.call_args_list

    def test_health_check_in_pipeline(self, r):
        with r.pipeline(transaction=False) as pipe:
            pipe.connection = pipe.connection_pool.get_connection("_")
            pipe.connection.next_health_check = 0
            with mock.patch.object(
                pipe.connection, "send_command", wraps=pipe.connection.send_command
            ) as m:
                responses = pipe.set("foo", "bar").get("foo").execute()
                m.assert_any_call("PING", check_health=False)
                assert responses == [True, b"bar"]

    def test_health_check_in_transaction(self, r):
        with r.pipeline(transaction=True) as pipe:
            pipe.connection = pipe.connection_pool.get_connection("_")
            pipe.connection.next_health_check = 0
            with mock.patch.object(
                pipe.connection, "send_command", wraps=pipe.connection.send_command
            ) as m:
                responses = pipe.set("foo", "bar").get("foo").execute()
                m.assert_any_call("PING", check_health=False)
                assert responses == [True, b"bar"]

    def test_health_check_in_watched_pipeline(self, r):
        r.set("foo", "bar")
        with r.pipeline(transaction=False) as pipe:
            pipe.connection = pipe.connection_pool.get_connection("_")
            pipe.connection.next_health_check = 0
            with mock.patch.object(
                pipe.connection, "send_command", wraps=pipe.connection.send_command
            ) as m:
                pipe.watch("foo")
                # the health check should be called when watching
                m.assert_called_with("PING", check_health=False)
                self.assert_interval_advanced(pipe.connection)
                assert pipe.get("foo") == b"bar"

                # reset the mock to clear the call list and schedule another
                # health check
                m.reset_mock()
                pipe.connection.next_health_check = 0

                pipe.multi()
                responses = pipe.set("foo", "not-bar").get("foo").execute()
                assert responses == [True, b"not-bar"]
                m.assert_any_call("PING", check_health=False)

    def test_health_check_in_pubsub_before_subscribe(self, r):
        "A health check happens before the first [p]subscribe"
        p = r.pubsub()
        p.connection = p.connection_pool.get_connection("_")
        p.connection.next_health_check = 0
        with mock.patch.object(
            p.connection, "send_command", wraps=p.connection.send_command
        ) as m:
            assert not p.subscribed
            p.subscribe("foo")
            # the connection is not yet in pubsub mode, so the normal
            # ping/pong within connection.send_command should check
            # the health of the connection
            m.assert_any_call("PING", check_health=False)
            self.assert_interval_advanced(p.connection)

            subscribe_message = wait_for_message(p)
            assert subscribe_message["type"] == "subscribe"

    def test_health_check_in_pubsub_after_subscribed(self, r):
        """
        Pubsub can handle a new subscribe when it's time to check the
        connection health
        """
        p = r.pubsub()
        p.connection = p.connection_pool.get_connection("_")
        p.connection.next_health_check = 0
        with mock.patch.object(
            p.connection, "send_command", wraps=p.connection.send_command
        ) as m:
            p.subscribe("foo")
            subscribe_message = wait_for_message(p)
            assert subscribe_message["type"] == "subscribe"
            self.assert_interval_advanced(p.connection)
            # because we weren't subscribed when sending the subscribe
            # message to 'foo', the connection's standard check_health ran
            # prior to subscribing.
            m.assert_any_call("PING", check_health=False)

            p.connection.next_health_check = 0
            m.reset_mock()

            p.subscribe("bar")
            # the second subscribe issues exactly only command (the subscribe)
            # and the health check is not invoked
            m.assert_called_once_with("SUBSCRIBE", "bar", check_health=False)

            # since no message has been read since the health check was
            # reset, it should still be 0
            assert p.connection.next_health_check == 0

            subscribe_message = wait_for_message(p)
            assert subscribe_message["type"] == "subscribe"
            assert wait_for_message(p) is None
            # now that the connection is subscribed, the pubsub health
            # check should have taken over and include the HEALTH_CHECK_MESSAGE
            m.assert_any_call("PING", p.HEALTH_CHECK_MESSAGE, check_health=False)
            self.assert_interval_advanced(p.connection)

    def test_health_check_in_pubsub_poll(self, r):
        """
        Polling a pubsub connection that's subscribed will regularly
        check the connection's health.
        """
        p = r.pubsub()
        p.connection = p.connection_pool.get_connection("_")
        with mock.patch.object(
            p.connection, "send_command", wraps=p.connection.send_command
        ) as m:
            p.subscribe("foo")
            subscribe_message = wait_for_message(p)
            assert subscribe_message["type"] == "subscribe"
            self.assert_interval_advanced(p.connection)

            # polling the connection before the health check interval
            # doesn't result in another health check
            m.reset_mock()
            next_health_check = p.connection.next_health_check
            assert wait_for_message(p) is None
            assert p.connection.next_health_check == next_health_check
            m.assert_not_called()

            # reset the health check and poll again
            # we should not receive a pong message, but the next_health_check
            # should be advanced
            p.connection.next_health_check = 0
            assert wait_for_message(p) is None
            m.assert_called_with("PING", p.HEALTH_CHECK_MESSAGE, check_health=False)
            self.assert_interval_advanced(p.connection)



from contextlib import asynccontextmanager as _asynccontextmanager
from typing import Union

import pytest
import pytest_asyncio
import valkey.asyncio as valkey
from tests.conftest import VALKEY_INFO
from valkey.asyncio import Sentinel
from valkey.asyncio.client import Monitor
from valkey.asyncio.connection import Connection, parse_url
from valkey.asyncio.retry import Retry
from valkey.backoff import NoBackoff

from .compat import mock


async def _get_info(valkey_url):
    client = valkey.Valkey.from_url(valkey_url)
    info = await client.info()
    await client.connection_pool.disconnect()
    return info


@pytest_asyncio.fixture(
    params=[
        pytest.param(
            (True,),
            marks=pytest.mark.skipif(
                'config.VALKEY_INFO["cluster_enabled"]', reason="cluster mode enabled"
            ),
        ),
        (False,),
    ],
    ids=[
        "single",
        "pool",
    ],
)
async def create_valkey(request):
    """Wrapper around valkey.create_valkey."""
    (single_connection,) = request.param

    teardown_clients = []

    async def client_factory(
        url: str = request.config.getoption("--valkey-url"),
        cls=valkey.Valkey,
        flushdb=True,
        **kwargs,
    ):
        if "protocol" not in url and kwargs.get("protocol") is None:
            kwargs["protocol"] = request.config.getoption("--protocol")

        cluster_mode = VALKEY_INFO["cluster_enabled"]
        if not cluster_mode:
            single = kwargs.pop("single_connection_client", False) or single_connection
            url_options = parse_url(url)
            url_options.update(kwargs)
            pool = valkey.ConnectionPool(**url_options)
            client = cls(connection_pool=pool)
        else:
            client = valkey.ValkeyCluster.from_url(url, **kwargs)
            await client.initialize()
            single = False
        if single:
            client = client.client()
            await client.initialize()

        async def teardown():
            if not cluster_mode:
                if flushdb and "username" not in kwargs:
                    try:
                        await client.flushdb()
                    except valkey.ConnectionError:
                        # handle cases where a test disconnected a client
                        # just manually retry the flushdb
                        await client.flushdb()
                await client.aclose()
                await client.connection_pool.disconnect()
            else:
                if flushdb:
                    try:
                        await client.flushdb(target_nodes="primaries")
                    except valkey.ConnectionError:
                        # handle cases where a test disconnected a client
                        # just manually retry the flushdb
                        await client.flushdb(target_nodes="primaries")
                await client.aclose()

        teardown_clients.append(teardown)
        return client

    yield client_factory

    for teardown in teardown_clients:
        await teardown()


@pytest_asyncio.fixture()
async def r(create_valkey):
    return await create_valkey()


@pytest_asyncio.fixture()
async def r2(create_valkey):
    """A second client for tests that need multiple"""
    return await create_valkey()


@pytest_asyncio.fixture()
async def decoded_r(create_valkey):
    return await create_valkey(decode_responses=True)


@pytest_asyncio.fixture()
async def sentinel_setup(local_cache, request):
    sentinel_ips = request.config.getoption("--sentinels")
    sentinel_endpoints = [
        (ip.strip(), int(port.strip()))
        for ip, port in (endpoint.split(":") for endpoint in sentinel_ips.split(","))
    ]
    kwargs = request.param.get("kwargs", {}) if hasattr(request, "param") else {}
    sentinel = Sentinel(
        sentinel_endpoints,
        socket_timeout=0.1,
        client_cache=local_cache,
        protocol=3,
        **kwargs,
    )
    yield sentinel
    for s in sentinel.sentinels:
        await s.aclose()


@pytest_asyncio.fixture()
async def master(request, sentinel_setup):
    master_service = request.config.getoption("--master-service")
    master = sentinel_setup.master_for(master_service)
    yield master
    await master.aclose()


def _gen_cluster_mock_resp(r, response):
    connection = mock.AsyncMock(spec=Connection)
    connection.retry = Retry(NoBackoff(), 0)
    connection.read_response.return_value = response
    connection._get_from_local_cache.return_value = None
    with mock.patch.object(r, "connection", connection):
        yield r


@pytest_asyncio.fixture()
async def mock_cluster_resp_ok(create_valkey, **kwargs):
    r = await create_valkey(**kwargs)
    for mocked in _gen_cluster_mock_resp(r, "OK"):
        yield mocked


@pytest_asyncio.fixture()
async def mock_cluster_resp_int(create_valkey, **kwargs):
    r = await create_valkey(**kwargs)
    for mocked in _gen_cluster_mock_resp(r, 2):
        yield mocked


@pytest_asyncio.fixture()
async def mock_cluster_resp_info(create_valkey, **kwargs):
    r = await create_valkey(**kwargs)
    response = (
        "cluster_state:ok\r\ncluster_slots_assigned:16384\r\n"
        "cluster_slots_ok:16384\r\ncluster_slots_pfail:0\r\n"
        "cluster_slots_fail:0\r\ncluster_known_nodes:7\r\n"
        "cluster_size:3\r\ncluster_current_epoch:7\r\n"
        "cluster_my_epoch:2\r\ncluster_stats_messages_sent:170262\r\n"
        "cluster_stats_messages_received:105653\r\n"
    )
    for mocked in _gen_cluster_mock_resp(r, response):
        yield mocked


@pytest_asyncio.fixture()
async def mock_cluster_resp_nodes(create_valkey, **kwargs):
    r = await create_valkey(**kwargs)
    response = (
        "c8253bae761cb1ecb2b61857d85dfe455a0fec8b 172.17.0.7:7006 "
        "slave aa90da731f673a99617dfe930306549a09f83a6b 0 "
        "1447836263059 5 connected\n"
        "9bd595fe4821a0e8d6b99d70faa660638a7612b3 172.17.0.7:7008 "
        "master - 0 1447836264065 0 connected\n"
        "aa90da731f673a99617dfe930306549a09f83a6b 172.17.0.7:7003 "
        "myself,master - 0 0 2 connected 5461-10922\n"
        "1df047e5a594f945d82fc140be97a1452bcbf93e 172.17.0.7:7007 "
        "slave 19efe5a631f3296fdf21a5441680f893e8cc96ec 0 "
        "1447836262556 3 connected\n"
        "4ad9a12e63e8f0207025eeba2354bcf4c85e5b22 172.17.0.7:7005 "
        "master - 0 1447836262555 7 connected 0-5460\n"
        "19efe5a631f3296fdf21a5441680f893e8cc96ec 172.17.0.7:7004 "
        "master - 0 1447836263562 3 connected 10923-16383\n"
        "fbb23ed8cfa23f17eaf27ff7d0c410492a1093d6 172.17.0.7:7002 "
        "master,fail - 1447829446956 1447829444948 1 disconnected\n"
    )
    for mocked in _gen_cluster_mock_resp(r, response):
        yield mocked


@pytest_asyncio.fixture()
async def mock_cluster_resp_slaves(create_valkey, **kwargs):
    r = await create_valkey(**kwargs)
    response = (
        "['1df047e5a594f945d82fc140be97a1452bcbf93e 172.17.0.7:7007 "
        "slave 19efe5a631f3296fdf21a5441680f893e8cc96ec 0 "
        "1447836789290 3 connected']"
    )
    for mocked in _gen_cluster_mock_resp(r, response):
        yield mocked


async def wait_for_command(
    client: valkey.Valkey, monitor: Monitor, command: str, key: Union[str, None] = None
):
    # issue a command with a key name that's local to this process.
    # if we find a command with our key before the command we're waiting
    # for, something went wrong
    if key is None:
        # generate key
        id_str = str(await client.client_id())
        key = f"__VALKEY-PY-{id_str}__"
    await client.get(key)
    while True:
        monitor_response = await monitor.next_command()
        if command in monitor_response["command"]:
            return monitor_response
        if key in monitor_response["command"]:
            return None


# python 3.6 doesn't have the asynccontextmanager decorator.  Provide it here.
class AsyncContextManager:
    def __init__(self, async_generator):
        self.gen = async_generator

    async def __aenter__(self):
        try:
            return await self.gen.__anext__()
        except StopAsyncIteration as err:
            raise RuntimeError("Pickles") from err

    async def __aexit__(self, exc_type, exc_inst, tb):
        if exc_type:
            await self.gen.athrow(exc_type, exc_inst, tb)
            return True
        try:
            await self.gen.__anext__()
        except StopAsyncIteration:
            return
        raise RuntimeError("More pickles")


def asynccontextmanager(func):
    return _asynccontextmanager(func)


# helpers to get the connection arguments for this run
@pytest.fixture()
def valkey_url(request):
    return request.config.getoption("--valkey-url")


@pytest.fixture()
def connect_args(request):
    url = request.config.getoption("--valkey-url")
    return parse_url(url)



from .cluster import READ_COMMANDS, AsyncValkeyClusterCommands, ValkeyClusterCommands
from .core import AsyncCoreCommands, CoreCommands
from .helpers import list_or_args
from .sentinel import AsyncSentinelCommands, SentinelCommands
from .valkeymodules import AsyncValkeyModuleCommands, ValkeyModuleCommands

__all__ = [
    "AsyncCoreCommands",
    "AsyncValkeyClusterCommands",
    "AsyncValkeyModuleCommands",
    "AsyncSentinelCommands",
    "CoreCommands",
    "READ_COMMANDS",
    "ValkeyClusterCommands",
    "ValkeyModuleCommands",
    "SentinelCommands",
    "list_or_args",
]



from ..helpers import nativestr


class BFInfo:
    capacity = None
    size = None
    filterNum = None
    insertedNum = None
    expansionRate = None

    def __init__(self, args):
        response = dict(zip(map(nativestr, args[::2]), args[1::2]))
        self.capacity = response["Capacity"]
        self.size = response["Size"]
        self.filterNum = response["Number of filters"]
        self.insertedNum = response["Number of items inserted"]
        self.expansionRate = response["Expansion rate"]

    def get(self, item):
        try:
            return self.__getitem__(item)
        except AttributeError:
            return None

    def __getitem__(self, item):
        return getattr(self, item)


class CFInfo:
    size = None
    bucketNum = None
    filterNum = None
    insertedNum = None
    deletedNum = None
    bucketSize = None
    expansionRate = None
    maxIteration = None

    def __init__(self, args):
        response = dict(zip(map(nativestr, args[::2]), args[1::2]))
        self.size = response["Size"]
        self.bucketNum = response["Number of buckets"]
        self.filterNum = response["Number of filters"]
        self.insertedNum = response["Number of items inserted"]
        self.deletedNum = response["Number of items deleted"]
        self.bucketSize = response["Bucket size"]
        self.expansionRate = response["Expansion rate"]
        self.maxIteration = response["Max iterations"]

    def get(self, item):
        try:
            return self.__getitem__(item)
        except AttributeError:
            return None

    def __getitem__(self, item):
        return getattr(self, item)


class CMSInfo:
    width = None
    depth = None
    count = None

    def __init__(self, args):
        response = dict(zip(map(nativestr, args[::2]), args[1::2]))
        self.width = response["width"]
        self.depth = response["depth"]
        self.count = response["count"]

    def __getitem__(self, item):
        return getattr(self, item)


class TopKInfo:
    k = None
    width = None
    depth = None
    decay = None

    def __init__(self, args):
        response = dict(zip(map(nativestr, args[::2]), args[1::2]))
        self.k = response["k"]
        self.width = response["width"]
        self.depth = response["depth"]
        self.decay = response["decay"]

    def __getitem__(self, item):
        return getattr(self, item)


class TDigestInfo:
    compression = None
    capacity = None
    merged_nodes = None
    unmerged_nodes = None
    merged_weight = None
    unmerged_weight = None
    total_compressions = None
    memory_usage = None

    def __init__(self, args):
        response = dict(zip(map(nativestr, args[::2]), args[1::2]))
        self.compression = response["Compression"]
        self.capacity = response["Capacity"]
        self.merged_nodes = response["Merged nodes"]
        self.unmerged_nodes = response["Unmerged nodes"]
        self.merged_weight = response["Merged weight"]
        self.unmerged_weight = response["Unmerged weight"]
        self.total_compressions = response["Total compressions"]
        self.memory_usage = response["Memory usage"]

    def get(self, item):
        try:
            return self.__getitem__(item)
        except AttributeError:
            return None

    def __getitem__(self, item):
        return getattr(self, item)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from valkey.asyncio.client import Pipeline, Valkey


def from_url(url, **kwargs):
    """
    Returns an active Valkey client generated from the given database URL.

    Will attempt to extract the database id from the path url fragment, if
    none is provided.
    """
    from valkey.asyncio.client import Valkey

    return Valkey.from_url(url, **kwargs)


class pipeline:
    def __init__(self, valkey_obj: "Valkey"):
        self.p: "Pipeline" = valkey_obj.pipeline()

    async def __aenter__(self) -> "Pipeline":
        return self.p

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.p.execute()
        del self.p
